"""Dependencies."""

import logging
from functools import cache
from typing import Annotated, Any, AsyncIterator

from fastapi import Depends, HTTPException, Request
from fastapi.security import HTTPBearer
from httpx import AsyncClient, HTTPStatusError
from keycloak import KeycloakOpenID
from langchain_openai import ChatOpenAI
from langgraph.checkpoint.base import BaseCheckpointSaver
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
from starlette.status import HTTP_401_UNAUTHORIZED

from neuroagent.agents import (
    BaseAgent,
    SimpleAgent,
    SimpleChatAgent,
)
from neuroagent.agents.base_agent import (
    AsyncPostgresSaverWithPrefix,
    AsyncSqliteSaverWithPrefix,
)
from neuroagent.app.app_utils import validate_project
from neuroagent.app.config import Settings
from neuroagent.app.routers.database.schemas import Threads
from neuroagent.cell_types import CellTypesMeta
from neuroagent.multi_agents import BaseMultiAgent, SupervisorMultiAgent
from neuroagent.tools import (
    BlueNaaSTool,
    ElectrophysFeatureTool,
    GetMEModelTool,
    GetMorphoTool,
    GetTracesTool,
    KGMorphoFeatureTool,
    LiteratureSearchTool,
    MorphologyFeatureTool,
    ResolveEntitiesTool,
)
from neuroagent.utils import RegionMeta, get_file_from_KG

logger = logging.getLogger(__name__)


class HTTPBearerDirect(HTTPBearer):
    """HTTPBearer class that returns directly the token in the call."""

    async def __call__(self, request: Request) -> str | None:  # type: ignore
        """Intercept the bearer token in the headers."""
        auth_credentials = await super().__call__(request)
        return auth_credentials.credentials if auth_credentials else None


auth = HTTPBearerDirect(auto_error=False)


@cache
def get_settings() -> Settings:
    """Load all parameters.

    Note that this function is cached and environment will be read just once.
    """
    logger.info("Reading the environment and instantiating settings")
    return Settings()


async def get_httpx_client(request: Request) -> AsyncIterator[AsyncClient]:
    """Manage the httpx client for the request."""
    client = AsyncClient(
        timeout=None,
        verify=False,
        headers={"x-request-id": request.headers["x-request-id"]},
    )
    yield client


def get_connection_string(
    settings: Annotated[Settings, Depends(get_settings)],
) -> str | None:
    """Get the db interacting class for chat agent."""
    if settings.db.prefix:
        connection_string = settings.db.prefix
        if settings.db.user and settings.db.password:
            # Add authentication.
            connection_string += (
                f"{settings.db.user}:{settings.db.password.get_secret_value()}@"
            )
        if settings.db.host:
            # Either in file DB or connect to remote host.
            connection_string += settings.db.host
        if settings.db.port:
            # Add the port if remote host.
            connection_string += f":{settings.db.port}"
        if settings.db.name:
            # Add database name if specified.
            connection_string += f"/{settings.db.name}"
        return connection_string
    else:
        return None


def get_engine(request: Request) -> AsyncEngine | None:
    """Get the SQL engine."""
    return request.app.state.engine


async def get_session(
    engine: Annotated[AsyncEngine | None, Depends(get_engine)],
) -> AsyncIterator[AsyncSession]:
    """Yield a session per request."""
    if not engine:
        raise HTTPException(
            status_code=500,
            detail={
                "detail": "Couldn't connect to the SQL DB.",
            },
        )
    async with AsyncSession(engine) as session:
        yield session


async def get_user_id(
    token: Annotated[str, Depends(auth)],
    settings: Annotated[Settings, Depends(get_settings)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> str:
    """Validate JWT token and returns user ID."""
    if settings.keycloak.validate_token and settings.keycloak.user_info_endpoint:
        try:
            response = await httpx_client.get(
                settings.keycloak.user_info_endpoint,
                headers={"Authorization": f"Bearer {token}"},
            )
            response.raise_for_status()
            user_info = response.json()
            return user_info["sub"]
        except HTTPStatusError:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Invalid token."
            )
    else:
        return "dev"


def get_kg_token(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str | None, Depends(auth)],
) -> str:
    """Get a Knowledge graph token using Keycloak."""
    if token:
        return token
    else:
        instance = KeycloakOpenID(
            server_url=settings.keycloak.server_url,
            realm_name=settings.keycloak.realm,
            client_id=settings.keycloak.client_id,
        )
        return instance.token(
            username=settings.keycloak.username,
            password=settings.keycloak.password.get_secret_value(),  # type: ignore
        )["access_token"]


def get_bluenaas_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> BlueNaaSTool:
    """Load BlueNaaS tool."""
    tool = BlueNaaSTool(
        metadata={
            "url": settings.tools.bluenaas.url,
            "token": token,
            "httpx_client": httpx_client,
        }
    )
    return tool


def get_literature_tool(
    token: Annotated[str, Depends(get_kg_token)],
    settings: Annotated[Settings, Depends(get_settings)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> LiteratureSearchTool:
    """Load literature tool."""
    tool = LiteratureSearchTool(
        metadata={
            "url": settings.tools.literature.url,
            "httpx_client": httpx_client,
            "token": token,
            "retriever_k": settings.tools.literature.retriever_k,
            "reranker_k": settings.tools.literature.reranker_k,
            "use_reranker": settings.tools.literature.use_reranker,
        }
    )
    return tool


def get_entities_resolver_tool(
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> ResolveEntitiesTool:
    """Load resolve brain region tool."""
    tool = ResolveEntitiesTool(
        metadata={
            "token": token,
            "httpx_client": httpx_client,
            "kg_sparql_url": settings.knowledge_graph.sparql_url,
            "kg_class_view_url": settings.knowledge_graph.class_view_url,
        }
    )
    return tool


def get_morpho_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> GetMorphoTool:
    """Load get morpho tool."""
    tool = GetMorphoTool(
        metadata={
            "url": settings.knowledge_graph.url,
            "token": token,
            "httpx_client": httpx_client,
            "search_size": settings.tools.morpho.search_size,
            "brainregion_path": settings.knowledge_graph.br_saving_path,
            "celltypes_path": settings.knowledge_graph.ct_saving_path,
        }
    )
    return tool


def get_kg_morpho_feature_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> KGMorphoFeatureTool:
    """Load knowledge graph tool."""
    tool = KGMorphoFeatureTool(
        metadata={
            "url": settings.knowledge_graph.url,
            "token": token,
            "httpx_client": httpx_client,
            "search_size": settings.tools.kg_morpho_features.search_size,
            "brainregion_path": settings.knowledge_graph.br_saving_path,
        }
    )
    return tool


def get_traces_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> GetTracesTool:
    """Load knowledge graph tool."""
    tool = GetTracesTool(
        metadata={
            "url": settings.knowledge_graph.url,
            "token": token,
            "httpx_client": httpx_client,
            "search_size": settings.tools.trace.search_size,
            "brainregion_path": settings.knowledge_graph.br_saving_path,
        }
    )
    return tool


def get_electrophys_feature_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> ElectrophysFeatureTool:
    """Load morphology features tool."""
    tool = ElectrophysFeatureTool(
        metadata={
            "url": settings.knowledge_graph.url,
            "token": token,
            "httpx_client": httpx_client,
        }
    )
    return tool


def get_morphology_feature_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> MorphologyFeatureTool:
    """Load morphology features tool."""
    tool = MorphologyFeatureTool(
        metadata={
            "url": settings.knowledge_graph.url,
            "token": token,
            "httpx_client": httpx_client,
        }
    )
    return tool


def get_me_model_tool(
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> GetMEModelTool:
    """Load get ME model tool."""
    tool = GetMEModelTool(
        metadata={
            "url": settings.knowledge_graph.url,
            "token": token,
            "httpx_client": httpx_client,
            "search_size": settings.tools.me_model.search_size,
            "brainregion_path": settings.knowledge_graph.br_saving_path,
            "celltypes_path": settings.knowledge_graph.ct_saving_path,
        }
    )
    return tool


def get_language_model(
    settings: Annotated[Settings, Depends(get_settings)],
) -> ChatOpenAI:
    """Get the language model."""
    logger.info(f"OpenAI selected. Loading model {settings.openai.model}.")
    return ChatOpenAI(
        model_name=settings.openai.model,
        temperature=settings.openai.temperature,
        openai_api_key=settings.openai.token.get_secret_value(),  # type: ignore
        max_tokens=settings.openai.max_tokens,
        seed=78,
        streaming=True,
    )


async def get_agent_memory(
    connection_string: Annotated[str | None, Depends(get_connection_string)],
) -> AsyncIterator[BaseCheckpointSaver[Any] | None]:
    """Get the agent checkpointer."""
    if connection_string:
        if connection_string.startswith("sqlite"):
            async with AsyncSqliteSaverWithPrefix.from_conn_string(
                connection_string
            ) as memory:
                await memory.setup()
                yield memory
                await memory.conn.close()

        elif connection_string.startswith("postgresql"):
            async with AsyncPostgresSaverWithPrefix.from_conn_string(
                connection_string
            ) as memory:
                await memory.setup()
                yield memory
                await memory.conn.close()
        else:
            raise HTTPException(
                status_code=500,
                detail={
                    "details": (
                        f"Database of type {connection_string.split(':')[0]} is not"
                        " supported."
                    )
                },
            )
    else:
        logger.warning("The SQL db_prefix needs to be set to use the SQL DB.")
        yield None


async def get_vlab_and_project(
    user_id: Annotated[str, Depends(get_user_id)],
    session: Annotated[AsyncSession, Depends(get_session)],
    request: Request,
    settings: Annotated[Settings, Depends(get_settings)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    token: Annotated[str, Depends(get_kg_token)],
) -> dict[str, str]:
    """Get the current vlab and project ID."""
    if "x-project-id" in request.headers and "x-virtual-lab-id" in request.headers:
        vlab_and_project = {
            "vlab_id": request.headers["x-virtual-lab-id"],
            "project_id": request.headers["x-project-id"],
        }
    elif not settings.keycloak.validate_token:
        vlab_and_project = {
            "vlab_id": "430108e9-a81d-4b13-b7b6-afca00195908",
            "project_id": "eff09ea1-be16-47f0-91b6-52a3ea3ee575",
        }
    else:
        thread_id = request.path_params.get("thread_id")
        thread = await session.get(Threads, (thread_id, user_id))
        if thread and thread.vlab_id and thread.project_id:
            vlab_and_project = {
                "vlab_id": thread.vlab_id,
                "project_id": thread.project_id,
            }
        else:
            raise HTTPException(
                status_code=404,
                detail="thread not found when trying to validate project ID.",
            )

    await validate_project(
        httpx_client=httpx_client,
        vlab_id=vlab_and_project["vlab_id"],
        project_id=vlab_and_project["project_id"],
        token=token,
        vlab_project_url=settings.virtual_lab.get_project_url,
    )
    return vlab_and_project


def get_agent(
    _: Annotated[None, Depends(get_vlab_and_project)],
    llm: Annotated[ChatOpenAI, Depends(get_language_model)],
    literature_tool: Annotated[LiteratureSearchTool, Depends(get_literature_tool)],
    entities_resolver_tool: Annotated[
        ResolveEntitiesTool, Depends(get_entities_resolver_tool)
    ],
    morpho_tool: Annotated[GetMorphoTool, Depends(get_morpho_tool)],
    morphology_feature_tool: Annotated[
        MorphologyFeatureTool, Depends(get_morphology_feature_tool)
    ],
    kg_morpho_feature_tool: Annotated[
        KGMorphoFeatureTool, Depends(get_kg_morpho_feature_tool)
    ],
    electrophys_feature_tool: Annotated[
        ElectrophysFeatureTool, Depends(get_electrophys_feature_tool)
    ],
    traces_tool: Annotated[GetTracesTool, Depends(get_traces_tool)],
    me_model_tool: Annotated[GetMEModelTool, Depends(get_me_model_tool)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> BaseAgent | BaseMultiAgent:
    """Get the generative question answering service."""
    if settings.agent.model == "multi":
        logger.info("Load multi-agent chat")
        tools_list = [
            ("literature", [literature_tool]),
            (
                "morphologies",
                [
                    entities_resolver_tool,
                    morpho_tool,
                    morphology_feature_tool,
                    kg_morpho_feature_tool,
                ],
            ),
            ("traces", [entities_resolver_tool, electrophys_feature_tool, traces_tool]),
        ]
        return SupervisorMultiAgent(llm=llm, agents=tools_list)  # type: ignore
    else:
        tools = [
            literature_tool,
            entities_resolver_tool,
            morpho_tool,
            morphology_feature_tool,
            kg_morpho_feature_tool,
            electrophys_feature_tool,
            traces_tool,
            me_model_tool,
        ]
        logger.info("Load simple agent")
        return SimpleAgent(llm=llm, tools=tools)  # type: ignore


def get_chat_agent(
    _: Annotated[None, Depends(get_vlab_and_project)],
    llm: Annotated[ChatOpenAI, Depends(get_language_model)],
    memory: Annotated[BaseCheckpointSaver[Any], Depends(get_agent_memory)],
    bluenaas_tool: Annotated[BlueNaaSTool, Depends(get_bluenaas_tool)],
    literature_tool: Annotated[LiteratureSearchTool, Depends(get_literature_tool)],
    entities_resolver_tool: Annotated[
        ResolveEntitiesTool, Depends(get_entities_resolver_tool)
    ],
    morpho_tool: Annotated[GetMorphoTool, Depends(get_morpho_tool)],
    morphology_feature_tool: Annotated[
        MorphologyFeatureTool, Depends(get_morphology_feature_tool)
    ],
    me_model_tool: Annotated[GetMEModelTool, Depends(get_me_model_tool)],
    kg_morpho_feature_tool: Annotated[
        KGMorphoFeatureTool, Depends(get_kg_morpho_feature_tool)
    ],
    electrophys_feature_tool: Annotated[
        ElectrophysFeatureTool, Depends(get_electrophys_feature_tool)
    ],
    traces_tool: Annotated[GetTracesTool, Depends(get_traces_tool)],
) -> BaseAgent:
    """Get the generative question answering service."""
    logger.info("Load simple chat")
    tools = [
        bluenaas_tool,
        literature_tool,
        entities_resolver_tool,
        me_model_tool,
        morpho_tool,
        morphology_feature_tool,
        kg_morpho_feature_tool,
        electrophys_feature_tool,
        traces_tool,
    ]
    return SimpleChatAgent(llm=llm, tools=tools, memory=memory)  # type: ignore


async def get_update_kg_hierarchy(
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    settings: Annotated[Settings, Depends(get_settings)],
    file_name: str = "brainregion.json",
) -> None:
    """Query file from KG and update the local hierarchy file."""
    file_url = f"<{settings.knowledge_graph.hierarchy_url}/brainregion>"
    KG_hierarchy = await get_file_from_KG(
        file_url=file_url,
        file_name=file_name,
        view_url=settings.knowledge_graph.sparql_url,
        token=token,
        httpx_client=httpx_client,
    )
    RegionMeta_temp = RegionMeta.from_KG_dict(KG_hierarchy)
    RegionMeta_temp.save_config(settings.knowledge_graph.br_saving_path)
    logger.info("Knowledge Graph Brain Regions Hierarchy file updated.")


async def get_cell_types_kg_hierarchy(
    token: Annotated[str, Depends(get_kg_token)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    settings: Annotated[Settings, Depends(get_settings)],
    file_name: str = "celltypes.json",
) -> None:
    """Query file from KG and update the local hierarchy file."""
    file_url = f"<{settings.knowledge_graph.hierarchy_url}/celltypes>"
    hierarchy = await get_file_from_KG(
        file_url=file_url,
        file_name=file_name,
        view_url=settings.knowledge_graph.sparql_url,
        token=token,
        httpx_client=httpx_client,
    )
    celltypesmeta = CellTypesMeta.from_dict(hierarchy)
    celltypesmeta.save_config(settings.knowledge_graph.ct_saving_path)
    logger.info("Knowledge Graph Cell Types Hierarchy file updated.")



"""Conversation related CRUD operations."""

import logging
from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException
from httpx import AsyncClient
from langchain_core.messages import AIMessage, HumanMessage
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from sqlalchemy import MetaData, select
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession

from neuroagent.app.app_utils import validate_project
from neuroagent.app.config import Settings
from neuroagent.app.dependencies import (
    get_agent_memory,
    get_engine,
    get_httpx_client,
    get_kg_token,
    get_session,
    get_settings,
    get_user_id,
)
from neuroagent.app.routers.database.schemas import (
    GetThreadsOutput,
    Threads,
    ThreadsRead,
    ThreadsUpdate,
)
from neuroagent.app.routers.database.sql import get_object

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/threads", tags=["Threads' CRUD"])


@router.post("/")
async def create_thread(
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    settings: Annotated[Settings, Depends(get_settings)],
    token: Annotated[str, Depends(get_kg_token)],
    session: Annotated[AsyncSession, Depends(get_session)],
    user_id: Annotated[str, Depends(get_user_id)],
    virtual_lab_id: str,
    project_id: str,
    title: str = "title",
) -> ThreadsRead:
    """Create thread.
    \f

    Parameters
    ----------
    session
        SQL session to communicate with the db.
    user_id
        ID of the current user.
    title
        Title of the thread to create.

    Returns
    -------
    thread_dict: {'thread_id': 'thread_name'}
        Conversation created.
    """  # noqa: D301, D400, D205
    # We first need to check if the combination thread/vlab/project is valid
    await validate_project(
        httpx_client=httpx_client,
        vlab_id=virtual_lab_id,
        project_id=project_id,
        token=token,
        vlab_project_url=settings.virtual_lab.get_project_url,
    )

    new_thread = Threads(
        user_sub=user_id, vlab_id=virtual_lab_id, title=title, project_id=project_id
    )
    session.add(new_thread)
    await session.commit()
    await session.refresh(new_thread)
    return ThreadsRead(**new_thread.__dict__)


@router.get("/")
async def get_threads(
    session: Annotated[AsyncSession, Depends(get_session)],
    user_id: Annotated[str, Depends(get_user_id)],
) -> list[ThreadsRead]:
    """Get threads for a user.
    \f

    Parameters
    ----------
    session
        SQL session to communicate with the db.
    user_id
        ID of the current user.

    Returns
    -------
    list[ThreadsRead]
        List the threads from the given user id.
    """  # noqa: D205, D301, D400
    query = select(Threads).where(Threads.user_sub == user_id)
    results = await session.execute(query)
    threads = results.all()
    return [ThreadsRead(**thread[0].__dict__) for thread in threads]


@router.get("/{thread_id}")
async def get_thread(
    _: Annotated[Threads, Depends(get_object)],
    memory: Annotated[AsyncSqliteSaver | None, Depends(get_agent_memory)],
    thread_id: str,
) -> list[GetThreadsOutput]:
    """Get thread.
    \f

    Parameters
    ----------
    _
        Thread object returned by SQLAlchemy
    memory
        Langgraph's checkpointer's instance.
    thread_id
        ID of the thread.

    Returns
    -------
    messages
        Conversation with messages.

    Raises
    ------
    HTTPException
        If the thread is not from the current user.
    """  # noqa: D301, D205, D400
    if memory is None:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Couldn't connect to the SQL DB.",
            },
        )
    config = RunnableConfig({"configurable": {"thread_id": thread_id}})
    messages = await memory.aget(config)
    if not messages:
        return []

    output: list[GetThreadsOutput] = []
    # Reconstruct the conversation. Also output message_id for other endpoints
    for message in messages["channel_values"]["messages"]:
        if isinstance(message, HumanMessage):
            output.append(
                GetThreadsOutput(
                    message_id=message.id,
                    entity="Human",
                    message=message.content,
                )
            )
        if isinstance(message, AIMessage) and message.content:
            output.append(
                GetThreadsOutput(
                    message_id=message.id,
                    entity="AI",
                    message=message.content,
                )
            )
    return output


@router.patch("/{thread_id}")
async def update_thread_title(
    thread: Annotated[Threads, Depends(get_object)],
    session: Annotated[AsyncSession, Depends(get_session)],
    thread_update: ThreadsUpdate,
) -> ThreadsRead:
    """Update thread.
    \f

    Parameters
    ----------
    thread
        Thread object returned by SQLAlchemy
    session
        SQL session
    thread_update
        Pydantic class containing the required updates

    Returns
    -------
    thread_return
        Updated thread instance
    """  # noqa: D205, D301, D400
    thread_data = thread_update.model_dump(exclude_unset=True)
    for key, value in thread_data.items():
        setattr(thread, key, value)
    session.add(thread)
    await session.commit()
    await session.refresh(thread)
    thread_return = ThreadsRead(**thread.__dict__)  # For mypy.
    return thread_return


@router.delete("/{thread_id}")
async def delete_thread(
    _: Annotated[Threads, Depends(get_object)],
    session: Annotated[AsyncSession, Depends(get_session)],
    engine: Annotated[AsyncEngine, Depends(get_engine)],
    thread_id: str,
) -> dict[str, str]:
    """Delete the specified thread.
    \f

    Parameters
    ----------
    _
        Thread object returned by SQLAlchemy
    session
        SQL session
    engine
        SQL engine
    thread_id
        ID of the relevant thread

    Returns
    -------
    Acknowledgement of the deletion
    """  # noqa: D205, D301, D400
    metadata = MetaData()
    async with engine.begin() as conn:
        await conn.run_sync(metadata.reflect)
    for table in metadata.tables.values():
        if "thread_id" not in table.c.keys():
            continue
        # Delete from the checkpoint table
        query = table.delete().where(table.c.thread_id == thread_id)
        await session.execute(query)

    await session.commit()
    return {"Acknowledged": "true"}



import pytest
from httpx import AsyncClient

from neuroagent.resolving import (
    es_resolve,
    escape_punctuation,
    resolve_query,
    sparql_exact_resolve,
    sparql_fuzzy_resolve,
)


@pytest.mark.asyncio
async def test_sparql_exact_resolve(httpx_mock, get_resolve_query_output):
    brain_region = "Thalamus"
    url = "http://fakeurl.com"
    mocked_response = get_resolve_query_output[0]
    httpx_mock.add_response(
        url=url,
        json=mocked_response,
    )
    response = await sparql_exact_resolve(
        query=brain_region,
        resource_type="nsg:BrainRegion",
        sparql_view_url=url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
    )
    assert response == [
        {
            "label": "Thalamus",
            "id": "http://api.brain-map.org/api/v2/data/Structure/549",
        }
    ]

    httpx_mock.reset()

    mtype = "Interneuron"
    mocked_response = get_resolve_query_output[1]
    httpx_mock.add_response(
        url=url,
        json=mocked_response,
    )
    response = await sparql_exact_resolve(
        query=mtype,
        resource_type="bmo:BrainCellType",
        sparql_view_url=url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
    )
    assert response == [
        {"label": "Interneuron", "id": "https://neuroshapes.org/Interneuron"}
    ]


@pytest.mark.asyncio
async def test_sparql_fuzzy_resolve(httpx_mock, get_resolve_query_output):
    brain_region = "Field"
    url = "http://fakeurl.com"
    mocked_response = get_resolve_query_output[2]
    httpx_mock.add_response(
        url=url,
        json=mocked_response,
    )
    response = await sparql_fuzzy_resolve(
        query=brain_region,
        resource_type="nsg:BrainRegion",
        sparql_view_url=url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
    )
    assert response == [
        {
            "label": "Field CA1",
            "id": "http://api.brain-map.org/api/v2/data/Structure/382",
        },
        {
            "label": "Field CA2",
            "id": "http://api.brain-map.org/api/v2/data/Structure/423",
        },
        {
            "label": "Field CA3",
            "id": "http://api.brain-map.org/api/v2/data/Structure/463",
        },
    ]
    httpx_mock.reset()

    mtype = "Interneu"
    mocked_response = get_resolve_query_output[3]
    httpx_mock.add_response(
        url=url,
        json=mocked_response,
    )
    response = await sparql_fuzzy_resolve(
        query=mtype,
        resource_type="bmo:BrainCellType",
        sparql_view_url=url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
    )
    assert response == [
        {"label": "Interneuron", "id": "https://neuroshapes.org/Interneuron"},
        {
            "label": "Hippocampus CA3 Oriens Interneuron",
            "id": "http://uri.interlex.org/base/ilx_0105044",
        },
        {
            "label": "Spinal Cord Ventral Horn Interneuron IA",
            "id": "http://uri.interlex.org/base/ilx_0110929",
        },
    ]


@pytest.mark.asyncio
async def test_es_resolve(httpx_mock, get_resolve_query_output):
    brain_region = "Auditory Cortex"
    mocked_response = get_resolve_query_output[4]
    httpx_mock.add_response(
        url="http://goodurl.com",
        json=mocked_response,
    )
    response = await es_resolve(
        query=brain_region,
        resource_type="nsg:BrainRegion",
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
        es_view_url="http://goodurl.com",
    )
    assert response == [
        {
            "label": "Cerebral cortex",
            "id": "http://api.brain-map.org/api/v2/data/Structure/688",
        },
        {
            "label": "Cerebellar cortex",
            "id": "http://api.brain-map.org/api/v2/data/Structure/528",
        },
        {
            "label": "Frontal pole, cerebral cortex",
            "id": "http://api.brain-map.org/api/v2/data/Structure/184",
        },
    ]
    httpx_mock.reset()

    mtype = "Ventral neuron"
    mocked_response = get_resolve_query_output[5]
    httpx_mock.add_response(
        url="http://goodurl.com",
        json=mocked_response,
    )
    response = await es_resolve(
        query=mtype,
        resource_type="bmo:BrainCellType",
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
        es_view_url="http://goodurl.com",
    )
    assert response == [
        {
            "label": "Ventral Tegmental Area Dopamine Neuron",
            "id": "http://uri.interlex.org/base/ilx_0112352",
        },
        {
            "label": "Spinal Cord Ventral Horn Motor Neuron Gamma",
            "id": "http://uri.interlex.org/base/ilx_0110943",
        },
        {
            "label": "Hypoglossal Nucleus Motor Neuron",
            "id": "http://uri.interlex.org/base/ilx_0105169",
        },
    ]


@pytest.mark.asyncio
async def test_resolve_query(httpx_mock, get_resolve_query_output):
    url = "http://terribleurl.com"
    class_view_url = "http://somewhatokurl.com"
    # Mock exact match to fail
    httpx_mock.add_response(
        url=url,
        json={
            "head": {"vars": ["subject", "predicate", "object", "context"]},
            "results": {"bindings": []},
        },
    )

    # Hit fuzzy match
    httpx_mock.add_response(
        url=url,
        json=get_resolve_query_output[2],
    )

    # Hit ES match
    httpx_mock.add_response(
        url=class_view_url,
        json=get_resolve_query_output[4],
    )
    response = await resolve_query(
        query="Field",
        resource_type="nsg:BrainRegion",
        sparql_view_url=url,
        es_view_url=class_view_url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
    )
    assert response == [
        {
            "label": "Field CA1",
            "id": "http://api.brain-map.org/api/v2/data/Structure/382",
        },
        {
            "label": "Field CA2",
            "id": "http://api.brain-map.org/api/v2/data/Structure/423",
        },
        {
            "label": "Field CA3",
            "id": "http://api.brain-map.org/api/v2/data/Structure/463",
        },
    ]
    httpx_mock.reset()

    httpx_mock.add_response(url=url, json=get_resolve_query_output[0])

    # Hit fuzzy match
    httpx_mock.add_response(
        url=url,
        json={
            "head": {"vars": ["subject", "predicate", "object", "context"]},
            "results": {"bindings": []},
        },
    )

    # Hit ES match
    httpx_mock.add_response(url=class_view_url, json={"hits": {"hits": []}})

    response = await resolve_query(
        query="Thalamus",
        resource_type="nsg:BrainRegion",
        sparql_view_url=url,
        es_view_url=class_view_url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
    )
    assert response == [
        {
            "label": "Thalamus",
            "id": "http://api.brain-map.org/api/v2/data/Structure/549",
        }
    ]
    httpx_mock.reset()
    httpx_mock.add_response(
        url=url,
        json={
            "head": {"vars": ["subject", "predicate", "object", "context"]},
            "results": {"bindings": []},
        },
    )

    # Hit fuzzy match
    httpx_mock.add_response(
        url=url,
        json={
            "head": {"vars": ["subject", "predicate", "object", "context"]},
            "results": {"bindings": []},
        },
    )

    # Hit ES match
    httpx_mock.add_response(
        url=class_view_url,
        json=get_resolve_query_output[4],
    )
    response = await resolve_query(
        query="Auditory Cortex",
        resource_type="nsg:BrainRegion",
        sparql_view_url=url,
        es_view_url=class_view_url,
        token="greattokenpleasedontexpire",
        httpx_client=AsyncClient(),
        search_size=3,
    )
    assert response == [
        {
            "label": "Cerebral cortex",
            "id": "http://api.brain-map.org/api/v2/data/Structure/688",
        },
        {
            "label": "Cerebellar cortex",
            "id": "http://api.brain-map.org/api/v2/data/Structure/528",
        },
        {
            "label": "Frontal pole, cerebral cortex",
            "id": "http://api.brain-map.org/api/v2/data/Structure/184",
        },
    ]


@pytest.mark.parametrize(
    "before,after",
    [
        ("this is a text", "this is a text"),
        ("this is text with punctuation!", "this is text with punctuation\\\\!"),
    ],
)
def test_escape_punctuation(before, after):
    assert after == escape_punctuation(before)


def test_failing_escape_punctuation():
    text = 15  # this is not a string
    with pytest.raises(TypeError) as e:
        escape_punctuation(text)
    assert e.value.args[0] == "Only accepting strings."



"""BlueNaaS single cell stimulation, simulation and synapse placement tool."""

import json
import logging
from typing import Annotated, Any, Literal

from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, ToolMessage
from langchain_core.tools import ToolException
from langgraph.prebuilt import InjectedState
from pydantic import BaseModel, Field

from neuroagent.tools.base_tool import BaseToolOutput, BasicTool

logger = logging.getLogger(__name__)


class RecordingLocation(BaseModel):
    """Configuration for the recording location in the simulation."""

    section: str = Field(default="soma[0]", description="Section to record from")
    offset: float = Field(
        default=0.5, ge=0, le=1, description="Offset in the section to record from"
    )


class InputBlueNaaS(BaseModel):
    """Inputs for the BlueNaaS single-neuron simulation."""

    me_model_id: str = Field(
        description=(
            "ID of the neuron model to be used in the simulation. The model ID can be"
            " fetched using the 'get-me-model-tool'."
        )
    )
    current_injection__inject_to: str = Field(
        default="soma[0]", description="Section to inject the current to."
    )
    current_injection__stimulus__stimulus_type: Literal[
        "current_clamp", "voltage_clamp", "conductance"
    ] = Field(default="current_clamp", description="Type of stimulus to be used.")
    current_injection__stimulus__stimulus_protocol: Literal[
        "ap_waveform", "idrest", "iv", "fire_pattern"
    ] = Field(default="ap_waveform", description="Stimulus protocol to be used.")

    current_injection__stimulus__amplitudes: list[float] = Field(
        default=[0.1],
        min_length=1,
        description="List of amplitudes for the stimulus",
    )
    record_from: list[RecordingLocation] = Field(
        default=[RecordingLocation()],
        description=(
            "List of sections to record from during the simulation. Each record"
            " configuration includes the section name and offset."
        ),
    )
    conditions__celsius: int = Field(
        default=34, ge=0, le=50, description="Temperature in celsius"
    )
    conditions__vinit: int = Field(default=-73, description="Initial voltage in mV")
    conditions__hypamp: int = Field(default=0, description="Holding current in nA")
    conditions__max_time: int = Field(
        default=100, le=3000, description="Maximum simulation time in ms"
    )
    conditions__time_step: float = Field(
        default=0.05, ge=0.001, le=10, description="Time step in ms"
    )
    conditions__seed: int = Field(default=100, description="Random seed")
    messages: Annotated[list[BaseMessage], InjectedState("messages")]


class BlueNaaSValidatedOutput(BaseToolOutput):
    """Should return a successful POST request."""

    status: Literal["success", "pending", "error"]


class BlueNaaSInvalidatedOutput(BaseModel):
    """Response to the user if the simulation has not been validated yet."""

    inputs: dict[str, Any]

    def __str__(self) -> str:
        """Format the response passed to the LLM."""
        return f"A simulation will be ran with the following inputs <json>{self.inputs}</json>. \n Please confirm that you are satisfied by the simulation parameters, or correct them accordingly."


class BlueNaaSTool(BasicTool):
    """Class defining the BlueNaaS tool."""

    name: str = "bluenaas-tool"
    description: str = """Runs a single-neuron simulation using the BlueNaaS service.
    Requires a "me_model_id" which must be fetched by get-me-model-tool.
    Optionally, the user can specify simulation parameters.
    The tool will always ask for config validation from the user before running.
    If the user mentions an existing configuration, it must always be passed in the tool first to get user's approval.
    Specify ALL of the parameters everytime you enter this tool.
    """
    metadata: dict[str, Any]
    args_schema: type[BaseModel] = InputBlueNaaS
    response_format: Literal["content", "content_and_artifact"] = "content_and_artifact"

    def _run(self) -> None:
        pass

    async def _arun(
        self,
        me_model_id: str,
        messages: Annotated[list[BaseMessage], InjectedState("messages")],
        current_injection__inject_to: str = "soma[0]",
        current_injection__stimulus__stimulus_type: Literal[
            "current_clamp", "voltage_clamp", "conductance"
        ] = "current_clamp",
        current_injection__stimulus__stimulus_protocol: Literal[
            "ap_waveform", "idrest", "iv", "fire_pattern"
        ] = "ap_waveform",
        current_injection__stimulus__amplitudes: list[float] | None = None,
        record_from: list[RecordingLocation] | None = None,
        conditions__celsius: int = 34,
        conditions__vinit: int = -73,
        conditions__hypamp: int = 0,
        conditions__max_time: int = 100,
        conditions__time_step: float = 0.05,
        conditions__seed: int = 100,
    ) -> tuple[BaseToolOutput | BaseModel, dict[str, bool]]:
        """Run the BlueNaaS tool."""
        logger.info("Running BlueNaaS tool")

        json_api = self.create_json_api(
            current_injection__inject_to=current_injection__inject_to,
            current_injection__stimulus__stimulus_type=current_injection__stimulus__stimulus_type,
            current_injection__stimulus__stimulus_protocol=current_injection__stimulus__stimulus_protocol,
            current_injection__stimulus__amplitudes=current_injection__stimulus__amplitudes,
            record_from=record_from,
            conditions__celsius=conditions__celsius,
            conditions__vinit=conditions__vinit,
            conditions__hypamp=conditions__hypamp,
            conditions__max_time=conditions__max_time,
            conditions__time_step=conditions__time_step,
            conditions__seed=conditions__seed,
        )

        if self.is_validated(messages, json_api):
            try:
                await self.metadata["httpx_client"].post(
                    url=self.metadata["url"],
                    params={"model_id": me_model_id},
                    headers={"Authorization": f'Bearer {self.metadata["token"]}'},
                    json=json_api,
                    timeout=5.0,
                )

                return BlueNaaSValidatedOutput(status="success"), {
                    "is_validated": False
                }

            except Exception as e:
                raise ToolException(str(e), self.name)
        else:
            return BlueNaaSInvalidatedOutput(inputs=json_api), {"is_validated": True}

    @staticmethod
    def create_json_api(
        current_injection__inject_to: str = "soma[0]",
        current_injection__stimulus__stimulus_type: Literal[
            "current_clamp", "voltage_clamp", "conductance"
        ] = "current_clamp",
        current_injection__stimulus__stimulus_protocol: Literal[
            "ap_waveform", "idrest", "iv", "fire_pattern"
        ] = "ap_waveform",
        current_injection__stimulus__amplitudes: list[float] | None = None,
        record_from: list[RecordingLocation] | None = None,
        conditions__celsius: int = 34,
        conditions__vinit: int = -73,
        conditions__hypamp: int = 0,
        conditions__max_time: int = 100,
        conditions__time_step: float = 0.05,
        conditions__seed: int = 100,
    ) -> dict[str, Any]:
        """Based on the simulation config, create a valid JSON for the API."""
        if not current_injection__stimulus__amplitudes:
            current_injection__stimulus__amplitudes = [0.1]
        if not record_from:
            record_from = [RecordingLocation()]
        json_api = {
            "currentInjection": {
                "injectTo": current_injection__inject_to,
                "stimulus": {
                    "stimulusType": current_injection__stimulus__stimulus_type,
                    "stimulusProtocol": current_injection__stimulus__stimulus_protocol,
                    "amplitudes": current_injection__stimulus__amplitudes,
                },
            },
            "recordFrom": [recording.model_dump() for recording in record_from],
            "conditions": {
                "celsius": conditions__celsius,
                "vinit": conditions__vinit,
                "hypamp": conditions__hypamp,
                "max_time": conditions__max_time,
                "time_step": conditions__time_step,
                "seed": conditions__seed,
            },
            "type": "single-neuron-simulation",
            "simulationDuration": conditions__max_time,
        }
        return json_api

    @staticmethod
    def is_validated(messages: list[BaseMessage], json_api: dict[str, Any]) -> bool:
        """Decide whether the current configuration has been validated by the user.

        Parameters
        ----------
        messages
            List of Langgraph messages extracted from the graph state.
        json_api
            Simulation configuration that the tool will run if it has been validated.

        Returns
        -------
        is_validated
            Boolean stating wether or not the configuration has been validated by the user.
        """
        # If it is the first time the tool is called in the conversation, need validation
        try:
            # Get the last bluenaas call
            last_bluenaas_call = next(
                (
                    message
                    for message in reversed(messages)
                    if isinstance(message, ToolMessage)
                    and message.name == "bluenaas-tool"
                )
            )
        except StopIteration:
            return False

        # Verify if the tool has been recently called to ask for validation
        # There has to be at least 3 messages in the state otherwise cannot be validated
        if len(messages) > 3:
            last_messages = messages[-4:-1]
            recently_validated = (
                isinstance(last_messages[-1], HumanMessage)  # Approval from the human
                and isinstance(
                    last_messages[-2], AIMessage
                )  # AI answering the human and asking for validation
                and isinstance(
                    last_messages[-3], ToolMessage
                )  # First tool call not validated
                and last_messages[-3].name == "bluenaas-tool"
            )
            # If it hasn't been recently validated, we need more validation
            if not recently_validated:
                return False
        # If there is not enough messages in the state to have a potential validation
        else:
            return False

        # If the previous simulation was started, ask for validation on the new one
        if not last_bluenaas_call.artifact.get("is_validated"):
            return False

        # Verify if the config has changed since previous call. Validate again if so
        old_config = json.loads(
            last_bluenaas_call.content.split("<json>")[-1]  # type: ignore
            .split("</json>")[0]
            .replace("'", '"')
        )
        return old_config == json_api



"""Simple agent."""

import logging
from typing import Any, AsyncIterator

from langchain_core.messages import AIMessage
from langgraph.prebuilt import create_react_agent
from pydantic import model_validator

from neuroagent.agents import AgentOutput, AgentStep, BaseAgent

logger = logging.getLogger(__name__)


class SimpleAgent(BaseAgent):
    """Simple Agent class."""

    @model_validator(mode="before")
    @classmethod
    def create_agent(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Instantiate the clients upon class creation."""
        # Initialise the agent with the tools
        data["agent"] = create_react_agent(
            model=data["llm"],
            tools=data["tools"],
            state_modifier="""You are a helpful assistant helping scientists with neuro-scientific questions.
                You must always specify in your answers from which brain regions the information is extracted.
                Do no blindly repeat the brain region requested by the user, use the output of the tools instead.""",
        )
        return data

    def run(self, query: str) -> Any:
        """Run the agent against a query.

        Parameters
        ----------
        query
            Query of the user

        Returns
        -------
            Processed output of the LLM
        """
        return self._process_output(self.agent.invoke({"messages": [("human", query)]}))

    async def arun(self, query: str) -> Any:
        """Run the agent against a query.

        Parameters
        ----------
        query
            Query of the user

        Returns
        -------
            Processed output of the LLM
        """
        result = await self.agent.ainvoke({"messages": [("human", query)]})
        return self._process_output(result)

    async def astream(self, query: str) -> AsyncIterator[str]:
        """Run the agent against a query in streaming way.

        Parameters
        ----------
        query
            Query of the user

        Returns
        -------
            Iterator streaming the processed output of the LLM
        """
        streamed_response = self.agent.astream_events({"messages": query}, version="v2")

        async for event in streamed_response:
            kind = event["event"]

            # newline everytime model starts streaming.
            if kind == "on_chat_model_start":
                yield "\n\n"  # These \n are to separate AI message from tools.
            # check for the model stream.
            if kind == "on_chat_model_stream":
                # check if we are calling the tools.
                data_chunk = event["data"]["chunk"]
                if "tool_calls" in data_chunk.additional_kwargs:
                    tool = data_chunk.additional_kwargs["tool_calls"]
                    if tool[0]["function"]["name"]:
                        yield (
                            f'\nCalling tool : {tool[0]["function"]["name"]} with'
                            " arguments : "
                        )  # This \n is for when there are multiple async tool calls.
                    if tool[0]["function"]["arguments"]:
                        yield tool[0]["function"]["arguments"]

                content = data_chunk.content
                if content:
                    yield content
        yield "\n"

    @staticmethod
    def _process_output(output: Any) -> AgentOutput:
        """Format the output.

        Parameters
        ----------
        output
            Raw output of the LLM

        Returns
        -------
            Unified output across different agent type.
        """
        # Gather tool name and arguments together
        agent_steps = [
            AgentStep(
                tool_name=tool_call["name"],
                arguments=tool_call["args"],
            )
            for step in output["messages"]
            if isinstance(step, AIMessage) and step.additional_kwargs
            for tool_call in step.tool_calls
        ]
        return AgentOutput(response=output["messages"][-1].content, steps=agent_steps)



"""Utilies for neuroagent."""

import json
import logging
import numbers
import re
from pathlib import Path
from typing import Any, Iterator

from httpx import AsyncClient

from neuroagent.schemas import KGMetadata

logger = logging.getLogger(__name__)


class RegionMeta:
    """Class holding the hierarchical region metadata.

    Typically, such information would be parsed from a `brain_regions.json`
    file.

    Parameters
    ----------
    background_id : int, optional
        Override the default ID for the background.
    """

    def __init__(self, background_id: int = 0) -> None:
        self.background_id = background_id
        self.root_id: int | None = None

        self.name_: dict[int, str] = {self.background_id: "background"}
        self.st_level: dict[int, int | None] = {self.background_id: None}

        self.parent_id: dict[int, int] = {self.background_id: background_id}
        self.children_ids: dict[int, list[int]] = {self.background_id: []}

    def children(self, region_id: int) -> tuple[int, ...]:
        """Get all child region IDs of a given region.

        Note that by children we mean only the direct children, much like
        by parent we only mean the direct parent. The cumulative quantities
        that span all generations are called ancestors and descendants.

        Parameters
        ----------
        region_id : int
            The region ID in question.

        Returns
        -------
        int
            The region ID of a child region.
        """
        return tuple(self.children_ids[region_id])

    def descendants(self, ids: int | list[int]) -> set[int]:
        """Find all descendants of given regions.

        The result is inclusive, i.e. the input region IDs will be
        included in the result.

        Parameters
        ----------
        ids : int or iterable of int
            A region ID or a collection of region IDs to collect
            descendants for.

        Returns
        -------
        set
            All descendant region IDs of the given regions, including the input
            regions themselves.
        """
        if isinstance(ids, numbers.Integral):
            unique_ids: set[int] = {ids}
        elif isinstance(ids, set):
            unique_ids = set(ids)

        def iter_descendants(region_id: int) -> Iterator[int]:
            """Iterate over all descendants of a given region ID.

            Parameters
            ----------
            region_id
                Integer representing the id of the region

            Returns
            -------
                Iterator with descendants of the region
            """
            yield region_id
            for child in self.children(region_id):
                yield child
                yield from iter_descendants(child)

        descendants = set()
        for id_ in unique_ids:
            descendants |= set(iter_descendants(id_))

        return descendants

    def save_config(self, json_file_path: str | Path) -> None:
        """Save the actual configuration in a json file.

        Parameters
        ----------
        json_file_path
            Path where to save the json file
        """
        to_save = {
            "root_id": self.root_id,
            "names": self.name_,
            "st_level": self.st_level,
            "parent_id": self.parent_id,
            "children_ids": self.children_ids,
        }
        with open(json_file_path, "w") as fs:
            fs.write(json.dumps(to_save))

    @classmethod
    def load_config(cls, json_file_path: str | Path) -> "RegionMeta":
        """Load a configuration in a json file and return a 'RegionMeta' instance.

        Parameters
        ----------
        json_file_path
            Path to the json file containing the brain region hierarchy

        Returns
        -------
            RegionMeta class with pre-loaded hierarchy
        """
        with open(json_file_path, "r") as fs:
            to_load = json.load(fs)

        # Needed to convert json 'str' keys to int.
        for k1 in to_load.keys():
            if not isinstance(to_load[k1], int):
                to_load[k1] = {int(k): v for k, v in to_load[k1].items()}

        self = cls()

        self.root_id = to_load["root_id"]
        self.name_ = to_load["names"]
        self.st_level = to_load["st_level"]
        self.parent_id = to_load["parent_id"]
        self.children_ids = to_load["children_ids"]

        return self

    @classmethod
    def from_KG_dict(cls, KG_hierarchy: dict[str, Any]) -> "RegionMeta":
        """Construct an instance from the json of the Knowledge Graph.

        Parameters
        ----------
        KG_hierarchy : dict
            The dictionary of the region hierarchy, provided by the KG.

        Returns
        -------
        region_meta : RegionMeta
            The initialized instance of this class.
        """
        self = cls()

        for brain_region in KG_hierarchy["defines"]:
            # Filter out wrong elements of the KG.
            if "identifier" in brain_region.keys():
                region_id = int(brain_region["identifier"])

                # Check if we are at root.
                if "isPartOf" not in brain_region.keys():
                    self.root_id = int(region_id)
                    self.parent_id[region_id] = self.background_id
                else:
                    # Strip url to only keep ID.
                    self.parent_id[region_id] = int(
                        brain_region["isPartOf"][0].rsplit("/")[-1]
                    )
                self.children_ids[region_id] = []

                self.name_[region_id] = brain_region["label"]

                if "st_level" not in brain_region.keys():
                    self.st_level[region_id] = None
                else:
                    self.st_level[region_id] = brain_region["st_level"]

        # Once every parents are set, we can deduce all childrens.
        for child_id, parent_id in self.parent_id.items():
            if parent_id is not None:
                self.children_ids[int(parent_id)].append(child_id)

        return self

    @classmethod
    def load_json(cls, json_path: Path | str) -> "RegionMeta":
        """Load the structure graph from a JSON file and create a Class instance.

        Parameters
        ----------
        json_path : str or pathlib.Path

        Returns
        -------
        RegionMeta
            The initialized instance of this class.
        """
        with open(json_path) as fh:
            KG_hierarchy = json.load(fh)

        return cls.from_KG_dict(KG_hierarchy)


def get_descendants_id(brain_region_id: str, json_path: str | Path) -> set[str]:
    """Get all descendant of a brain region id.

    Parameters
    ----------
    brain_region_id
        Brain region ID to find descendants for.
    json_path
        Path to the json file containing the BR hierarchy

    Returns
    -------
        Set of descendants of a brain region
    """
    # Split a brain region ID of the form "http://api.brain-map.org/api/v2/data/Structure/123" into base + id.
    id_base, _, brain_region_str = brain_region_id.rpartition("/")
    try:
        # Convert the id into an int
        brain_region_int = int(brain_region_str)

        # Get the descendant ids of this BR (as int).
        region_meta = RegionMeta.load_config(json_path)
        hierarchy = region_meta.descendants(brain_region_int)

        # Recast the descendants into the form "http://api.brain-map.org/api/v2/data/Structure/123"
        hierarchy_ids = {f"{id_base}/{h}" for h in hierarchy}
    except ValueError:
        logger.info(
            f"The brain region {brain_region_id} didn't end with an int. Returning only"
            " the parent one."
        )
        hierarchy_ids = {brain_region_id}
    except IOError:
        logger.warning(f"The file {json_path} doesn't exist.")
        hierarchy_ids = {brain_region_id}

    return hierarchy_ids


async def get_file_from_KG(
    file_url: str,
    file_name: str,
    view_url: str,
    token: str,
    httpx_client: AsyncClient,
) -> dict[str, Any]:
    """Get json file for brain region / cell types from the KG.

    Parameters
    ----------
    file_url
        URL of the view containing the potential file
    file_name
        Name of the file to download
    view_url
        URL of the sparql view where to send the request to get the file url
    token
        Token used to access the knowledge graph
    httpx_client
        AsyncClient to send requests

    Returns
    -------
        Json contained in the downloaded file
    """
    sparql_query = """
    PREFIX schema: <http://schema.org/>

    SELECT DISTINCT ?file_url
    WHERE {{
        {file_url} schema:distribution ?json_distribution .
        ?json_distribution schema:name "{file_name}" ;
                        schema:contentUrl ?file_url .
    }}
    LIMIT 1""".format(file_url=file_url, file_name=file_name)
    try:
        file_response = None

        # Get the url of the relevant file
        url_response = await httpx_client.post(
            url=view_url,
            content=sparql_query,
            headers={
                "Content-Type": "text/plain",
                "Accept": "application/sparql-results+json",
                "Authorization": f"Bearer {token}",
            },
        )

        # Download the file
        file_response = await httpx_client.get(
            url=url_response.json()["results"]["bindings"][0]["file_url"]["value"],
            headers={
                "Accept": "*/*",
                "Authorization": f"Bearer {token}",
            },
        )

        return file_response.json()

    except ValueError:
        # Issue with KG
        if url_response.status_code != 200:
            raise ValueError(
                f"Could not find the file url, status code : {url_response.status_code}"
            )
        # File not found
        elif file_response:
            raise ValueError(
                f"Could not find the file, status code : {file_response.status_code}"
            )
        else:
            # Issue when downloading the file
            raise ValueError("url_response did not return a Json.")
    except IndexError:
        # No file url found
        raise IndexError("No file url was found.")
    except KeyError:
        # Json has weird format
        raise KeyError("Incorrect json format.")


def is_lnmc(contributors: list[dict[str, Any]]) -> bool:
    """Extract contributor affiliation out of the contributors."""
    lnmc_contributors = {
        "https://www.grid.ac/institutes/grid.5333.6",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/yshi",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/jyi",
        "https://bbp.epfl.ch/neurosciencegraph/data/664380c8-5a22-4974-951c-68ca78c0b1f1",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/perin",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/rajnish",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/ajaquier",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/gevaert",
        "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/kanari",
    }
    for contributor in contributors:
        if "@id" in contributor and contributor["@id"] in lnmc_contributors:
            return True

    return False


async def get_kg_data(
    object_id: str,
    httpx_client: AsyncClient,
    url: str,
    token: str,
    preferred_format: str,
) -> tuple[bytes, KGMetadata]:
    """Download any knowledge graph object.

    Parameters
    ----------
    object_id
        ID of the object to which the file is attached
    httpx_client
        AsyncClient to send the request
    url
        URL of the KG view where the object is located
    token
        Token used to access the knowledge graph
    preferred_format
        Extension of the file to download

    Returns
    -------
        Tuple containing the file's content and the associated metadata

    Raises
    ------
    ValueError
        If the object ID is not found the knowledge graph.
    """
    # Extract the id from the specified input (useful for rewoo)
    extracted_id = re.findall(pattern=r"https?://\S+[a-zA-Z0-9]", string=object_id)
    if not extracted_id:
        raise ValueError(f"The provided ID ({object_id}) is not valid.")
    else:
        object_id = extracted_id[0]

    # Create ES query to retrieve the object in KG
    query = {
        "size": 1,
        "track_total_hits": True,
        "query": {
            "bool": {
                "must": [
                    {
                        "term": {
                            "@id.keyword": object_id,
                        }
                    }
                ]
            }
        },
    }

    # Retrieve the object of interest from KG
    response = await httpx_client.post(
        url=url,
        headers={"Authorization": f"Bearer {token}"},
        json=query,
    )

    if response.status_code != 200 or len(response.json()["hits"]["hits"]) == 0:
        raise ValueError(f"We did not find the object {object_id} you are asking")

    # Get the metadata of the object
    response_data = response.json()["hits"]["hits"][0]["_source"]

    # Ensure we got the expected object
    if response_data["@id"] != object_id:
        raise ValueError(f"We did not find the object {object_id} you are asking")

    metadata: dict[str, Any] = dict()
    metadata["brain_region"] = response_data["brainRegion"]["label"]
    distributions = response_data["distribution"]

    # Extract the format of the file
    has_preferred_format = [
        i
        for i, dis in enumerate(distributions)
        if dis["encodingFormat"] == f"application/{preferred_format}"
    ]

    # Set the file extension accordingly if preferred format found
    if len(has_preferred_format) > 0:
        chosen_dist = distributions[has_preferred_format[0]]
        metadata["file_extension"] = preferred_format
    else:
        chosen_dist = distributions[0]
        metadata["file_extension"] = chosen_dist["encodingFormat"].split("/")[1]
        logger.info(
            "The format you specified was not available."
            f" {metadata['file_extension']} was chosen instead."
        )

    # Check if the object has been added by the LNMC lab (useful for traces)
    if "contributors" in response_data:
        metadata["is_lnmc"] = is_lnmc(response_data["contributors"])

    # Download the file
    url = chosen_dist["contentUrl"]
    content_response = await httpx_client.get(
        url=url,
        headers={"Authorization": f"Bearer {token}"},
    )

    # Return its content and the associated metadata
    object_content = content_response.content
    return object_content, KGMetadata(**metadata)



"""Test utility functions."""

import json
from pathlib import Path

import pytest
from httpx import AsyncClient

from neuroagent.schemas import KGMetadata
from neuroagent.utils import (
    RegionMeta,
    get_descendants_id,
    get_file_from_KG,
    get_kg_data,
    is_lnmc,
)


@pytest.mark.parametrize(
    "brain_region_id,expected_descendants",
    [
        ("brain-region-id/68", {"brain-region-id/68"}),
        (
            "another-brain-region-id/985",
            {
                "another-brain-region-id/320",
                "another-brain-region-id/648",
                "another-brain-region-id/844",
                "another-brain-region-id/882",
                "another-brain-region-id/943",
                "another-brain-region-id/985",
                "another-brain-region-id/3718675619",
                "another-brain-region-id/1758306548",
            },
        ),
        (
            "another-brain-region-id/369",
            {
                "another-brain-region-id/450",
                "another-brain-region-id/369",
                "another-brain-region-id/1026",
                "another-brain-region-id/854",
                "another-brain-region-id/577",
                "another-brain-region-id/625",
                "another-brain-region-id/945",
                "another-brain-region-id/1890964946",
                "another-brain-region-id/3693772975",
            },
        ),
        (
            "another-brain-region-id/178",
            {
                "another-brain-region-id/316",
                "another-brain-region-id/178",
                "another-brain-region-id/300",
                "another-brain-region-id/1043765183",
            },
        ),
        ("brain-region-id/not-a-int", {"brain-region-id/not-a-int"}),
    ],
)
def test_get_descendants(brain_region_id, expected_descendants, brain_region_json_path):
    descendants = get_descendants_id(brain_region_id, json_path=brain_region_json_path)
    assert expected_descendants == descendants


def test_get_descendants_errors(brain_region_json_path):
    brain_region_id = "does-not-exits/1111111111"
    with pytest.raises(KeyError):
        get_descendants_id(brain_region_id, json_path=brain_region_json_path)


def test_RegionMeta_from_KG_dict():
    with open(
        Path(__file__).parent / "data" / "KG_brain_regions_hierarchy_test.json"
    ) as fh:
        KG_hierarchy = json.load(fh)

    RegionMeta_test = RegionMeta.from_KG_dict(KG_hierarchy)

    # check names.
    assert RegionMeta_test.name_[1] == "Tuberomammillary nucleus, ventral part"
    assert (
        RegionMeta_test.name_[2]
        == "Superior colliculus, motor related, intermediate gray layer"
    )
    assert RegionMeta_test.name_[3] == "Primary Motor Cortex"

    # check parents / childrens.
    assert RegionMeta_test.parent_id[1] == 2
    assert RegionMeta_test.parent_id[2] == 0
    assert RegionMeta_test.parent_id[3] == 2
    assert RegionMeta_test.children_ids[1] == []
    assert RegionMeta_test.children_ids[2] == [1, 3]
    assert RegionMeta_test.children_ids[3] == []


def test_RegionMeta_save_load(tmp_path: Path):
    # load fake file from KG
    with open(
        Path(__file__).parent / "data" / "KG_brain_regions_hierarchy_test.json"
    ) as fh:
        KG_hierarchy = json.load(fh)

    RegionMeta_test = RegionMeta.from_KG_dict(KG_hierarchy)

    # save / load file.
    json_file = tmp_path / "test.json"
    RegionMeta_test.save_config(json_file)
    RegionMeta_test.load_config(json_file)

    # check names.
    assert RegionMeta_test.name_[1] == "Tuberomammillary nucleus, ventral part"
    assert (
        RegionMeta_test.name_[2]
        == "Superior colliculus, motor related, intermediate gray layer"
    )
    assert RegionMeta_test.name_[3] == "Primary Motor Cortex"

    # check parents / childrens.
    assert RegionMeta_test.parent_id[1] == 2
    assert RegionMeta_test.parent_id[2] == 0
    assert RegionMeta_test.parent_id[3] == 2
    assert RegionMeta_test.children_ids[1] == []
    assert RegionMeta_test.children_ids[2] == [1, 3]
    assert RegionMeta_test.children_ids[3] == []


def test_RegionMeta_load_real_file(brain_region_json_path):
    RegionMeta_test = RegionMeta.load_config(brain_region_json_path)

    # check root.
    assert RegionMeta_test.root_id == 997
    assert RegionMeta_test.parent_id[997] == 0

    # check some names / st_levels.
    assert RegionMeta_test.name_[123] == "Koelliker-Fuse subnucleus"
    assert RegionMeta_test.name_[78] == "middle cerebellar peduncle"
    assert RegionMeta_test.st_level[55] == 10

    # check some random parents / childrens.
    assert RegionMeta_test.parent_id[12] == 165
    assert RegionMeta_test.parent_id[78] == 752
    assert RegionMeta_test.parent_id[700] == 88
    assert RegionMeta_test.parent_id[900] == 840
    assert RegionMeta_test.children_ids[12] == []
    assert RegionMeta_test.children_ids[23] == []
    assert RegionMeta_test.children_ids[670] == [2260827822, 3562104832]
    assert RegionMeta_test.children_ids[31] == [1053, 179, 227, 39, 48, 572, 739]


@pytest.mark.asyncio
async def test_get_file_from_KG_errors(httpx_mock):
    file_url = "http://fake_url.com"
    file_name = "fake_file"
    view_url = "http://fake_url_view.com"
    token = "fake_token"
    client = AsyncClient()

    # first response from KG is not a json
    httpx_mock.add_response(url=view_url, text="not a json")

    with pytest.raises(ValueError) as not_json:
        await get_file_from_KG(
            file_url=file_url,
            file_name=file_name,
            view_url=view_url,
            token=token,
            httpx_client=client,
        )
    assert not_json.value.args[0] == "url_response did not return a Json."

    # no file url found in the KG
    httpx_mock.add_response(
        url=view_url, json={"head": {"vars": ["file_url"]}, "results": {"bindings": []}}
    )

    with pytest.raises(IndexError) as not_found:
        await get_file_from_KG(
            file_url=file_url,
            file_name=file_name,
            view_url=view_url,
            token=token,
            httpx_client=client,
        )
    assert not_found.value.args[0] == "No file url was found."

    httpx_mock.reset()
    # no file found corresponding to file_url
    test_file_url = "http://test_url.com"
    json_response = {
        "head": {"vars": ["file_url"]},
        "results": {
            "bindings": [{"file_url": {"type": "uri", "value": test_file_url}}]
        },
    }

    httpx_mock.add_response(url=view_url, json=json_response)
    httpx_mock.add_response(url=test_file_url, status_code=401)

    with pytest.raises(ValueError) as not_found:
        await get_file_from_KG(
            file_url=file_url,
            file_name=file_name,
            view_url=view_url,
            token=token,
            httpx_client=client,
        )
    assert not_found.value.args[0] == "Could not find the file, status code : 401"

    # Problem finding the file url
    httpx_mock.add_response(url=view_url, status_code=401)

    with pytest.raises(ValueError) as not_found:
        await get_file_from_KG(
            file_url=file_url,
            file_name=file_name,
            view_url=view_url,
            token=token,
            httpx_client=client,
        )
    assert not_found.value.args[0] == "Could not find the file url, status code : 401"


@pytest.mark.asyncio
async def test_get_file_from_KG(httpx_mock):
    file_url = "http://fake_url"
    file_name = "fake_file"
    view_url = "http://fake_url"
    token = "fake_token"
    test_file_url = "http://test_url"
    client = AsyncClient()

    json_response_url = {
        "head": {"vars": ["file_url"]},
        "results": {
            "bindings": [{"file_url": {"type": "uri", "value": test_file_url}}]
        },
    }
    with open(
        Path(__file__).parent / "data" / "KG_brain_regions_hierarchy_test.json"
    ) as fh:
        json_response_file = json.load(fh)

    httpx_mock.add_response(url=view_url, json=json_response_url)
    httpx_mock.add_response(url=test_file_url, json=json_response_file)

    response = await get_file_from_KG(
        file_url=file_url,
        file_name=file_name,
        view_url=view_url,
        token=token,
        httpx_client=client,
    )

    assert response == json_response_file


@pytest.mark.asyncio
async def test_get_kg_data_errors(httpx_mock):
    url = "http://fake_url"
    token = "fake_token"
    client = AsyncClient()

    # First failure: invalid object_id
    with pytest.raises(ValueError) as invalid_object_id:
        await get_kg_data(
            object_id="invalid_object_id",
            httpx_client=client,
            url=url,
            token=token,
            preferred_format="preferred_format",
        )

    assert (
        invalid_object_id.value.args[0]
        == "The provided ID (invalid_object_id) is not valid."
    )

    # Second failure: Number of hits = 0
    httpx_mock.add_response(url=url, json={"hits": {"hits": []}})

    with pytest.raises(ValueError) as no_hits:
        await get_kg_data(
            object_id="https://object-id",
            httpx_client=client,
            url=url,
            token=token,
            preferred_format="preferred_format",
        )

    assert (
        no_hits.value.args[0]
        == "We did not find the object https://object-id you are asking"
    )

    # Third failure: Wrong object id
    httpx_mock.add_response(
        url=url, json={"hits": {"hits": [{"_source": {"@id": "wrong-object-id"}}]}}
    )

    with pytest.raises(ValueError) as wrong_object_id:
        await get_kg_data(
            object_id="https://object-id",
            httpx_client=client,
            url=url,
            token=token,
            preferred_format="preferred_format",
        )

    assert (
        wrong_object_id.value.args[0]
        == "We did not find the object https://object-id you are asking"
    )


@pytest.mark.httpx_mock(can_send_already_matched_responses=True)
@pytest.mark.asyncio
async def test_get_kg_data(httpx_mock):
    url = "http://fake_url"
    token = "fake_token"
    client = AsyncClient()
    preferred_format = "txt"
    object_id = "https://object-id"

    response_json = {
        "hits": {
            "hits": [
                {
                    "_source": {
                        "@id": object_id,
                        "distribution": [
                            {
                                "encodingFormat": f"application/{preferred_format}",
                                "contentUrl": "http://content-url-txt",
                            }
                        ],
                        "contributors": [
                            {
                                "@id": "https://www.grid.ac/institutes/grid.5333.6",
                            }
                        ],
                        "brainRegion": {
                            "@id": "http://api.brain-map.org/api/v2/data/Structure/252",
                            "idLabel": (
                                "http://api.brain-map.org/api/v2/data/Structure/252|Dorsal"
                                " auditory area, layer 5"
                            ),
                            "identifier": (
                                "http://api.brain-map.org/api/v2/data/Structure/252"
                            ),
                            "label": "Dorsal auditory area, layer 5",
                        },
                    }
                }
            ]
        }
    }
    httpx_mock.add_response(
        url=url,
        json=response_json,
    )

    httpx_mock.add_response(
        url="http://content-url-txt",
        content=b"this is the txt content",
    )

    # Response with preferred format
    object_content, metadata = await get_kg_data(
        object_id="https://object-id",
        httpx_client=client,
        url=url,
        token=token,
        preferred_format=preferred_format,
    )

    assert isinstance(object_content, bytes)
    assert isinstance(metadata, KGMetadata)
    assert metadata.file_extension == "txt"
    assert metadata.is_lnmc is True

    # Response without preferred format
    object_content, reader = await get_kg_data(
        object_id="https://object-id",
        httpx_client=client,
        url=url,
        token=token,
        preferred_format="no_preferred_format_available",
    )

    assert isinstance(object_content, bytes)
    assert isinstance(metadata, KGMetadata)
    assert metadata.file_extension == "txt"
    assert metadata.is_lnmc is True


@pytest.mark.parametrize(
    "contributors,expected_bool",
    [
        (
            [
                {
                    "@id": "https://www.grid.ac/institutes/grid.5333.6",
                    "@type": ["http://schema.org/Organization"],
                    "label": "École Polytechnique Fédérale de Lausanne",
                }
            ],
            True,
        ),
        (
            [
                {
                    "@id": "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/gevaert",
                    "@type": ["http://schema.org/Person"],
                    "affiliation": "École Polytechnique Fédérale de Lausanne",
                }
            ],
            True,
        ),
        (
            [
                {},
                {
                    "@id": "https://bbp.epfl.ch/nexus/v1/realms/bbp/users/kanari",
                    "@type": ["http://schema.org/Person"],
                    "affiliation": "École Polytechnique Fédérale de Lausanne",
                },
            ],
            True,
        ),
        (
            [
                {
                    "@id": "wrong-id",
                    "@type": ["http://schema.org/Person"],
                    "affiliation": "Another school",
                }
            ],
            False,
        ),
    ],
)
def test_is_lnmc(contributors, expected_bool):
    assert is_lnmc(contributors) is expected_bool



"""Morphology features tool."""

import logging
from typing import Any, Type

import neurom
import numpy as np
from langchain_core.tools import ToolException
from neurom.io.utils import load_morphology
from pydantic import BaseModel, Field

from neuroagent.tools.base_tool import BaseToolOutput, BasicTool
from neuroagent.utils import get_kg_data

logger = logging.getLogger(__name__)


class InputNeuroM(BaseModel):
    """Inputs of the NeuroM API."""

    morphology_id: str = Field(
        description=(
            "ID of the morphology of interest. A morphology ID is an HTTP(S) link, it"
            " should therefore match the following regex pattern:"
            r" 'https?://\S+[a-zA-Z0-9]'"
        )
    )


class MorphologyFeatureOutput(BaseToolOutput):
    """Output schema for the neurom tool."""

    brain_region: str
    feature_dict: dict[str, Any]


class MorphologyFeatureTool(BasicTool):
    """Class defining the morphology feature retrieval logic."""

    name: str = "morpho-features-tool"
    description: str = """Given a morphology ID, fetch data about the features of the morphology. You need to know a morphology ID to use this tool and they can only come from the 'get-morpho-tool'. Therefore this tool should only be used if you already called the 'knowledge-graph-tool'.
    Here is an exhaustive list of features that can be retrieved with this tool:
    Soma radius, Soma surface area, Number of neurites, Number of sections, Number of sections per neurite, Section lengths, Segment lengths, Section radial distance, Section path distance, Local bifurcation angles, Remote bifurcation angles."""
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputNeuroM

    def _run(self) -> None:
        """Not implemented yet."""
        pass

    async def _arun(self, morphology_id: str) -> list[MorphologyFeatureOutput]:
        """Give features about morphology.

        Parameters
        ----------
        morphology_id
            ID of the morphology of interest

        Returns
        -------
            Dict containing feature_name: value.
        """
        logger.info(f"Entering morphology feature tool. Inputs: {morphology_id=}")
        try:
            # Download the .swc file describing the morphology from the KG
            morphology_content, metadata = await get_kg_data(
                object_id=morphology_id,
                httpx_client=self.metadata["httpx_client"],
                url=self.metadata["url"],
                token=self.metadata["token"],
                preferred_format="swc",
            )

            # Extract the features from it
            features = self.get_features(morphology_content, metadata.file_extension)
            return [
                MorphologyFeatureOutput(
                    brain_region=metadata.brain_region, feature_dict=features
                )
            ]
        except Exception as e:
            raise ToolException(str(e), self.name)

    def get_features(self, morphology_content: bytes, reader: str) -> dict[str, Any]:
        """Get features from a morphology.

        Parameters
        ----------
        morphology_content
            Bytes of the file containing the morphology info (comes from the KG)
        reader
            type of file (i.e. its extension)

        Returns
        -------
            Dict containing feature_name: value.
        """
        # Load the morphology
        morpho = load_morphology(morphology_content.decode(), reader=reader)

        # Compute soma radius and soma surface area
        features = {
            "soma_radius [µm]": neurom.get("soma_radius", morpho),
            "soma_surface_area [µm^2]": neurom.get("soma_surface_area", morpho),
        }

        # Prepare a list of features that have a unique value (no statistics)
        f1 = [
            ("number_of_neurites", "Number of neurites"),
            ("number_of_sections", "Number of sections"),
            ("number_of_sections_per_neurite", "Number of sections per neurite"),
        ]

        # For each neurite type, compute the above features
        for neurite_type in neurom.NEURITE_TYPES:
            for get_name, name in f1:
                features[f"{name} ({neurite_type.name})"] = neurom.get(
                    get_name, morpho, neurite_type=neurite_type
                )

        # Prepare a list of features that are defined by statistics
        f2 = [
            ("section_lengths", "Section lengths [µm]"),
            ("segment_lengths", "Segment lengths [µm]"),
            ("section_radial_distances", "Section radial distance [µm]"),
            ("section_path_distances", "Section path distance [µm]"),
            ("local_bifurcation_angles", "Local bifurcation angles [˚]"),
            ("remote_bifurcation_angles", "Remote bifurcation angles [˚]"),
        ]

        # For each neurite, compute the feature values and return their statistics
        for neurite_type in neurom.NEURITE_TYPES:
            for get_name, name in f2:
                try:
                    array = neurom.get(get_name, morpho, neurite_type=neurite_type)
                    if len(array) == 0:
                        continue
                    features[f"{name} ({neurite_type.name})"] = self.get_stats(array)
                except (IndexError, ValueError):
                    continue
        return features

    @staticmethod
    def get_stats(array: list[int | float]) -> dict[str, int | np.float64]:
        """Get summary stats for the array.

        Parameters
        ----------
        array
            Array of feature's statistics of a morphology

        Returns
        -------
            Dict with length, mean, sum, standard deviation, min and max of data
        """
        return {
            "len": len(array),
            "mean": np.mean(array),
            "sum": np.sum(array),
            "std": np.std(array),
            "min": np.min(array),
            "max": np.max(array),
        }



"""Agents."""

from neuroagent.agents.base_agent import AgentOutput, AgentStep, BaseAgent
from neuroagent.agents.simple_agent import SimpleAgent
from neuroagent.agents.simple_chat_agent import SimpleChatAgent

__all__ = [
    "AgentOutput",
    "AgentStep",
    "BaseAgent",
    "SimpleChatAgent",
    "SimpleAgent",
]



"""Supervisor multi-agent."""

import functools
import logging
import operator
from typing import Annotated, Any, AsyncIterator, Hashable, Sequence, TypedDict

from langchain_core.messages import AIMessage, BaseMessage, HumanMessage
from langchain_core.output_parsers.openai_functions import JsonOutputFunctionsParser
from langchain_core.prompts import (
    ChatPromptTemplate,
    MessagesPlaceholder,
    PromptTemplate,
)
from langchain_core.runnables import RunnableConfig
from langgraph.graph import END, START, StateGraph
from langgraph.graph.graph import CompiledGraph
from langgraph.prebuilt import create_react_agent
from pydantic import ConfigDict, model_validator

from neuroagent.agents import AgentOutput, AgentStep
from neuroagent.multi_agents.base_multi_agent import BaseMultiAgent

logger = logging.getLogger(__file__)


class AgentState(TypedDict):
    """Base class for agent state."""

    messages: Annotated[Sequence[BaseMessage], operator.add]
    next: str  # noqa: A003


class SupervisorMultiAgent(BaseMultiAgent):
    """Base class for multi agents."""

    summarizer: Any

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    @classmethod
    def create_main_agent(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Instantiate the clients upon class creation."""
        logger.info("Creating main agent, supervisor and all the agents with tools.")
        system_prompt = (
            "You are a supervisor tasked with managing a conversation between the"
            " following workers: {members}. Given the following user request,"
            " respond with the worker to act next. Each worker will perform a"
            " task and respond with their results and status. When finished,"
            " respond with FINISH."
        )
        agents_list = [elem[0] for elem in data["agents"]]
        logger.info(f"List of agents name: {agents_list}")

        options = ["FINISH"] + agents_list
        function_def = {
            "name": "route",
            "description": "Select the next role.",
            "parameters": {
                "title": "routeSchema",
                "type": "object",
                "properties": {
                    "next": {
                        "title": "Next",
                        "anyOf": [
                            {"enum": options},
                        ],
                    }
                },
                "required": ["next"],
            },
        }
        prompt = ChatPromptTemplate.from_messages(
            [
                ("system", system_prompt),
                MessagesPlaceholder(variable_name="messages"),
                (
                    "system",
                    (
                        "Given the conversation above, who should act next?"
                        " Or should we FINISH? Select one of: {options}"
                    ),
                ),
            ]
        ).partial(options=str(options), members=", ".join(agents_list))
        data["main_agent"] = (
            prompt
            | data["llm"].bind_functions(
                functions=[function_def], function_call="route"
            )
            | JsonOutputFunctionsParser()
        )
        data["summarizer"] = (
            PromptTemplate.from_template(
                """You are an helpful assistant. Here is the question of the user: {question}.
            And here are the results of the different tools used to answer: {responses}.
            You must always specify in your answers from which brain regions the information is extracted.
            Do no blindly repeat the brain region requested by the user, use the output of the tools instead.
            Keep all information that can be useful for the user such as the ids and links.
            Please formulate a complete response to give to the user ONLY based on the results.
            """
            )
            | data["llm"]
        )

        return data

    @staticmethod
    async def agent_node(
        state: AgentState, agent: CompiledGraph, name: str
    ) -> dict[str, Any]:
        """Run the agent node."""
        logger.info(f"Start running the agent: {name}")
        result = await agent.ainvoke(state)

        agent_steps = [
            AgentStep(
                tool_name=step.tool_calls[0]["name"],
                arguments=step.tool_calls[0]["args"],
            )
            for step in result["messages"]
            if isinstance(step, AIMessage) and step.additional_kwargs
        ]

        return {
            "messages": [
                AIMessage(
                    content=result["messages"][-1].content,
                    name=name,
                    additional_kwargs={"steps": agent_steps},
                )
            ]
        }

    async def summarizer_node(self, state: AgentState) -> dict[str, Any]:
        """Create summarizer node."""
        logger.info("Entering the summarizer node")
        question = state["messages"][0].content
        responses = " \n".join([mes.content for mes in state["messages"][1:]])  # type: ignore
        result = await self.summarizer.ainvoke(
            {"question": question, "responses": responses}
        )
        return {
            "messages": [
                HumanMessage(
                    content=result.content,
                    name="summarizer",
                )
            ]
        }

    def create_graph(self) -> CompiledGraph:
        """Create graph."""
        workflow = StateGraph(AgentState)

        # Create nodes
        for agent_name, tools_list in self.agents:
            agent = create_react_agent(model=self.llm, tools=tools_list)
            node = functools.partial(self.agent_node, agent=agent, name=agent_name)
            workflow.add_node(agent_name, node)

        # Supervisor node
        workflow.add_node("Supervisor", self.main_agent)

        # Summarizer node
        summarizer_agent = functools.partial(self.summarizer_node)
        workflow.add_node("Summarizer", summarizer_agent)

        # Create edges
        for agent_name, _ in self.agents:
            workflow.add_edge(agent_name, "Supervisor")

        conditional_map: dict[Hashable, str] = {k[0]: k[0] for k in self.agents}
        conditional_map["FINISH"] = "Summarizer"
        workflow.add_conditional_edges(
            "Supervisor",
            lambda x: x["next"],
            conditional_map,
        )
        workflow.add_edge(START, "Supervisor")
        workflow.add_edge("Summarizer", END)
        graph = workflow.compile()
        return graph

    def run(self, query: str, thread_id: str) -> AgentOutput:
        """Run graph against a query."""
        graph = self.create_graph()
        config = RunnableConfig(configurable={"thread_id": thread_id})
        res = graph.invoke(
            input={"messages": [HumanMessage(content=query)]}, config=config
        )
        return self._process_output(res)

    async def arun(self, query: str, thread_id: str) -> AgentOutput:
        """Arun method of the service."""
        graph = self.create_graph()
        config = RunnableConfig(configurable={"thread_id": thread_id})
        res = await graph.ainvoke(
            input={"messages": [HumanMessage(content=query)]}, config=config
        )
        return self._process_output(res)

    async def astream(self, query: str, thread_id: str) -> AsyncIterator[str]:  # type: ignore
        """Astream method of the service."""
        graph = self.create_graph()
        config = RunnableConfig(configurable={"thread_id": thread_id})
        async for chunk in graph.astream(
            input={"messages": [HumanMessage(content=query)]}, config=config
        ):
            if "Supervisor" in chunk.keys() and chunk["Supervisor"]["next"] != "FINISH":
                yield f'\nCalling agent : {chunk["Supervisor"]["next"]}\n'
            else:
                values = [i for i in chunk.values()]  # noqa: C416
                if "messages" in values[0]:
                    yield f'\n {values[0]["messages"][0].content}'

    @staticmethod
    def _process_output(output: Any) -> AgentOutput:
        """Format the output."""
        agent_steps = []
        for message in output["messages"][1:]:
            if "steps" in message.additional_kwargs:
                agent_steps.extend(message.additional_kwargs["steps"])
        return AgentOutput(response=output["messages"][-1].content, steps=agent_steps)



"""Middleware."""

from typing import Any, Callable

from fastapi import Request, Response

from neuroagent.app.dependencies import get_settings


async def strip_path_prefix(
    request: Request, call_next: Callable[[Any], Any]
) -> Response:
    """Optionally strip a prefix from a request path.

    Parameters
    ----------
    request
        Request sent by the user.
    call_next
        Function executed to get the output of the endpoint.

    Returns
    -------
    response: Response of the request after potentially stripping prefix from path and applying other middlewares
    """
    if request.base_url in (
        "http://testserver/",
        "http://test/",
    ) and "healthz" not in str(request.url):
        settings = request.app.dependency_overrides[get_settings]()
    else:
        settings = get_settings()
    prefix = settings.misc.application_prefix
    if prefix is not None and len(prefix) > 0 and request.url.path.startswith(prefix):
        new_path = request.url.path[len(prefix) :]
        scope = request.scope
        scope["path"] = new_path
        request = Request(scope, request.receive)
    return await call_next(request)



"""Multi-agents."""

from neuroagent.multi_agents.base_multi_agent import BaseMultiAgent
from neuroagent.multi_agents.supervisor_multi_agent import SupervisorMultiAgent

__all__ = [
    "BaseMultiAgent",
    "SupervisorMultiAgent",
]



"""Configuration."""

import os
import pathlib
from typing import Literal, Optional

from dotenv import dotenv_values
from pydantic import BaseModel, ConfigDict, SecretStr, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class SettingsAgent(BaseModel):
    """Agent setting."""

    model: Literal["simple", "multi"] = "simple"

    model_config = ConfigDict(frozen=True)


class SettingsDB(BaseModel):
    """DB settings for retrieving history."""

    prefix: str | None = None
    user: str | None = None
    password: SecretStr | None = None
    host: str | None = None
    port: str | None = None
    name: str | None = None

    model_config = ConfigDict(frozen=True)


class SettingsKeycloak(BaseModel):
    """Class retrieving keycloak info for authorization."""

    issuer: str = "https://openbluebrain.com/auth/realms/SBO"
    validate_token: bool = False
    # Useful only for service account (dev)
    client_id: str | None = None
    username: str | None = None
    password: SecretStr | None = None

    model_config = ConfigDict(frozen=True)

    @property
    def token_endpoint(self) -> str | None:
        """Define the token endpoint."""
        if self.validate_token:
            return f"{self.issuer}/protocol/openid-connect/token"
        else:
            return None

    @property
    def user_info_endpoint(self) -> str | None:
        """Define the user_info endpoint."""
        if self.validate_token:
            return f"{self.issuer}/protocol/openid-connect/userinfo"
        else:
            return None

    @property
    def server_url(self) -> str:
        """Server url."""
        return self.issuer.split("/auth")[0] + "/auth/"

    @property
    def realm(self) -> str:
        """Realm."""
        return self.issuer.rpartition("/realms/")[-1]


class SettingsLiterature(BaseModel):
    """Literature search API settings."""

    url: str
    retriever_k: int = 500
    use_reranker: bool = True
    reranker_k: int = 8

    model_config = ConfigDict(frozen=True)


class SettingsTrace(BaseModel):
    """Trace tool settings."""

    search_size: int = 10

    model_config = ConfigDict(frozen=True)


class SettingsKGMorpho(BaseModel):
    """KG Morpho settings."""

    search_size: int = 3

    model_config = ConfigDict(frozen=True)


class SettingsGetMorpho(BaseModel):
    """Get Morpho settings."""

    search_size: int = 10

    model_config = ConfigDict(frozen=True)


class SettingsGetMEModel(BaseModel):
    """Get ME Model settings."""

    search_size: int = 10

    model_config = ConfigDict(frozen=True)


class SettingsBlueNaaS(BaseModel):
    """BlueNaaS settings."""

    url: str = "https://openbluebrain.com/api/bluenaas/simulation/single-neuron/run"
    model_config = ConfigDict(frozen=True)


class SettingsKnowledgeGraph(BaseModel):
    """Knowledge graph API settings."""

    base_url: str
    download_hierarchy: bool = False
    br_saving_path: pathlib.Path | str = str(
        pathlib.Path(__file__).parent / "data" / "brainregion_hierarchy.json"
    )
    ct_saving_path: pathlib.Path | str = str(
        pathlib.Path(__file__).parent / "data" / "celltypes_hierarchy.json"
    )
    model_config = ConfigDict(frozen=True)

    @property
    def url(self) -> str:
        """Knowledge graph search url."""
        return f"{self.base_url}/search/query/"

    @property
    def sparql_url(self) -> str:
        """Knowledge graph view for sparql query."""
        return f"{self.base_url}/views/neurosciencegraph/datamodels/https%3A%2F%2Fbluebrain.github.io%2Fnexus%2Fvocabulary%2FdefaultSparqlIndex/sparql"

    @property
    def class_view_url(self) -> str:
        """Knowledge graph view for ES class query."""
        return f"{self.base_url}/views/neurosciencegraph/datamodels/https%3A%2F%2Fbbp.epfl.ch%2Fneurosciencegraph%2Fdata%2Fviews%2Fes%2Fdataset/_search"

    @property
    def hierarchy_url(self) -> str:
        """Knowledge graph url for brainregion/celltype files."""
        return "http://bbp.epfl.ch/neurosciencegraph/ontologies/core"


class SettingsVlab(BaseModel):
    """Virtual lab endpoint settings."""

    get_project_url: str = (
        "https://openbluebrain.com/api/virtual-lab-manager/virtual-labs"
    )
    model_config = ConfigDict(frozen=True)


class SettingsTools(BaseModel):
    """Database settings."""

    literature: SettingsLiterature
    bluenaas: SettingsBlueNaaS = SettingsBlueNaaS()
    morpho: SettingsGetMorpho = SettingsGetMorpho()
    trace: SettingsTrace = SettingsTrace()
    kg_morpho_features: SettingsKGMorpho = SettingsKGMorpho()
    me_model: SettingsGetMEModel = SettingsGetMEModel()

    model_config = ConfigDict(frozen=True)


class SettingsOpenAI(BaseModel):
    """OpenAI settings."""

    token: Optional[SecretStr] = None
    model: str = "gpt-4o-mini"
    temperature: float = 0
    max_tokens: Optional[int] = None

    model_config = ConfigDict(frozen=True)


class SettingsLogging(BaseModel):
    """Metadata settings."""

    level: Literal["debug", "info", "warning", "error", "critical"] = "info"
    external_packages: Literal["debug", "info", "warning", "error", "critical"] = (
        "warning"
    )

    model_config = ConfigDict(frozen=True)


class SettingsMisc(BaseModel):
    """Other settings."""

    application_prefix: str = ""
    # list is not hashable, the cors_origins have to be provided as a string with
    # comma separated entries, i.e. "value_1, value_2, ..."
    cors_origins: str = ""

    model_config = ConfigDict(frozen=True)


class Settings(BaseSettings):
    """All settings."""

    tools: SettingsTools
    knowledge_graph: SettingsKnowledgeGraph
    agent: SettingsAgent = SettingsAgent()  # has no required
    db: SettingsDB = SettingsDB()  # has no required
    openai: SettingsOpenAI = SettingsOpenAI()  # has no required
    logging: SettingsLogging = SettingsLogging()  # has no required
    keycloak: SettingsKeycloak = SettingsKeycloak()  # has no required
    virtual_lab: SettingsVlab = SettingsVlab()  # has no required
    misc: SettingsMisc = SettingsMisc()  # has no required

    model_config = SettingsConfigDict(
        env_file=".env",
        env_prefix="NEUROAGENT_",
        env_nested_delimiter="__",
        frozen=True,
    )

    @model_validator(mode="after")
    def check_consistency(self) -> "Settings":
        """Check if consistent.

        ATTENTION: Do not put model validators into the child settings. The
        model validator is run during instantiation.

        """
        # If you don't enforce keycloak auth, you need a way to communicate with the APIs the tools leverage
        if not self.keycloak.password and not self.keycloak.validate_token:
            raise ValueError(
                "Need an auth method for subsequent APIs called by the tools."
            )

        return self


# Load the remaining variables into the environment
# Necessary for things like SSL_CERT_FILE
config = dotenv_values()
for k, v in config.items():
    if k.lower().startswith("neuroagent_"):
        continue
    if v is None:
        continue
    os.environ[k] = os.environ.get(k, v)  # environment has precedence



"""SQL related functions."""

from typing import Annotated

from fastapi import Depends, HTTPException
from fastapi.security import HTTPBasic
from sqlalchemy.ext.asyncio import AsyncSession

from neuroagent.app.dependencies import get_session, get_user_id
from neuroagent.app.routers.database.schemas import Threads

security = HTTPBasic()


async def get_object(
    session: Annotated[AsyncSession, Depends(get_session)],
    thread_id: str,
    user_id: Annotated[str, Depends(get_user_id)],
) -> Threads:
    """Get an SQL object. Also useful to correlate user_id to thread_id.

    Parameters
    ----------
    session
        Session object connected to the SQL instance.
    thread_id
        ID of the thread, provided by the user.
    user_id
        ID of the user.

    Returns
    -------
    object
        Relevant row of the relevant table in the SQL DB.
    """
    sql_object = await session.get(Threads, (thread_id, user_id))
    if not sql_object:
        raise HTTPException(
            status_code=404,
            detail={
                "detail": "Thread not found.",
            },
        )
    return sql_object



"""Base multi-agent."""

from abc import ABC, abstractmethod
from typing import Any, AsyncIterator

from langchain.chat_models.base import BaseChatModel
from pydantic import BaseModel, ConfigDict

from neuroagent.agents import AgentOutput
from neuroagent.tools.base_tool import BasicTool


class BaseMultiAgent(BaseModel, ABC):
    """Base class for multi agents."""

    llm: BaseChatModel
    main_agent: Any
    agents: list[tuple[str, list[BasicTool]]]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> AgentOutput:
        """Run method of the service."""

    @abstractmethod
    async def arun(self, *args: Any, **kwargs: Any) -> AgentOutput:
        """Arun method of the service."""

    @abstractmethod
    async def astream(self, *args: Any, **kwargs: Any) -> AsyncIterator[str]:
        """Astream method of the service."""

    @staticmethod
    @abstractmethod
    def _process_output(*args: Any, **kwargs: Any) -> AgentOutput:
        """Format the output."""



"""Base tool (to handle errors)."""

import json
import logging
from typing import Any, Literal

from langchain_core.tools import BaseTool, ToolException
from pydantic import BaseModel, ValidationError, model_validator

logger = logging.getLogger(__name__)

EtypesLiteral = Literal[
    "bSTUT",
    "dSTUT",
    "bNAC",
    "cSTUT",
    "dNAC",
    "bAC",
    "cIR",
    "cAC",
    "cACint",
    "bIR",
    "cNAC",
    "cAD",
    "cADpyr",
    "cAD_ltb",
    "cNAD_ltb",
    "cAD_noscltb",
    "cNAD_noscltb",
    "dAD_ltb",
    "dNAD_ltb",
]
ETYPE_IDS = {
    "bSTUT": "http://uri.interlex.org/base/ilx_0738200",
    "dSTUT": "http://uri.interlex.org/base/ilx_0738202",
    "bNAC": "http://uri.interlex.org/base/ilx_0738203",
    "cSTUT": "http://uri.interlex.org/base/ilx_0738198",
    "dNAC": "http://uri.interlex.org/base/ilx_0738205",
    "bAC": "http://uri.interlex.org/base/ilx_0738199",
    "cIR": "http://uri.interlex.org/base/ilx_0738204",
    "cAC": "http://uri.interlex.org/base/ilx_0738197",
    "cACint": "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
    "bIR": "http://uri.interlex.org/base/ilx_0738206",
    "cNAC": "http://uri.interlex.org/base/ilx_0738201",
    "cAD": "http://uri.interlex.org/base/ilx_0738207",  # Both are the same id, what's the purpose ?
    "cADpyr": "http://uri.interlex.org/base/ilx_0738207",  # Both are the same id, what's the purpose ?
    "cAD_ltb": "http://uri.interlex.org/base/ilx_0738255",
    "cNAD_ltb": "http://uri.interlex.org/base/ilx_0738254",
    "cAD_noscltb": "http://uri.interlex.org/base/ilx_0738250",
    "cNAD_noscltb": "http://uri.interlex.org/base/ilx_0738249",
    "dAD_ltb": "http://uri.interlex.org/base/ilx_0738258",
    "dNAD_ltb": "http://uri.interlex.org/base/ilx_0738256",
}


def process_validation_error(error: ValidationError) -> str:
    """Handle validation errors when tool inputs are wrong."""
    error_list = []
    name = error.title
    # We have to iterate, in case there are multiple errors.
    try:
        for err in error.errors():
            if err["type"] == "literal_error":
                error_list.append(
                    {
                        "Validation error": (
                            f'Wrong value: provided {err["input"]} for input'
                            f' {err["loc"][0]}. Try again and change this problematic'
                            " input."
                        )
                    }
                )
            elif err["type"] == "missing":
                error_list.append(
                    {
                        "Validation error": (
                            f'Missing input : {err["loc"][0]}. Try again and add this'
                            " input."
                        )
                    }
                )
            else:
                error_list.append(
                    {"Validation error": f'{err["loc"][0]}. {err["msg"]}'}
                )

    except (KeyError, IndexError) as e:
        error_list.append({"Validation error": f"Error in {name} : {str(e)}"})
        logger.error(
            "UNTREATED ERROR !! PLEASE CONTACT ML TEAM AND FOWARD THEM THE REQUEST !!"
        )

    logger.warning(f"VALIDATION ERROR: Wrong input in {name}. {error_list}")

    return json.dumps(error_list)


def process_tool_error(error: ToolException) -> str:
    """Handle errors inside tools."""
    logger.warning(
        f"TOOL ERROR: Error in tool {error.args[1]}. Error: {str(error.args[0])}"
    )
    dict_output = {error.args[1]: error.args[0]}
    return json.dumps(dict_output)


class BasicTool(BaseTool):
    """Basic class for tools."""

    name: str = "base"
    description: str = "Base tool from which regular tools should inherit."

    @model_validator(mode="before")
    @classmethod
    def handle_errors(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Instantiate the clients upon class creation."""
        data["handle_validation_error"] = process_validation_error
        data["handle_tool_error"] = process_tool_error
        return data


class BaseToolOutput(BaseModel):
    """Base class for tool outputs."""

    def __repr__(self) -> str:
        """Representation method."""
        return self.model_dump_json()



"""Base agent."""

from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator

from langchain.chat_models.base import BaseChatModel
from langchain_core.tools import BaseTool
from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver
from langgraph.checkpoint.serde.base import SerializerProtocol
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from pydantic import BaseModel, ConfigDict


class AgentStep(BaseModel):
    """Class for agent decision steps."""

    tool_name: str
    arguments: dict[str, Any] | str


class AgentOutput(BaseModel):
    """Class for agent response."""

    response: str
    steps: list[AgentStep]


class BaseAgent(BaseModel, ABC):
    """Base class for services."""

    llm: BaseChatModel
    tools: list[BaseTool]
    agent: Any

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> AgentOutput:
        """Run method of the service."""

    @abstractmethod
    async def arun(self, *args: Any, **kwargs: Any) -> AgentOutput:
        """Arun method of the service."""

    @abstractmethod
    def astream(self, *args: Any, **kwargs: Any) -> AsyncIterator[str]:
        """Astream method of the service."""

    @staticmethod
    @abstractmethod
    def _process_output(*args: Any, **kwargs: Any) -> AgentOutput:
        """Format the output."""


class AsyncSqliteSaverWithPrefix(AsyncSqliteSaver):
    """Wrapper around the AsyncSqliteSaver that accepts a connection string with prefix."""

    @classmethod
    @asynccontextmanager
    async def from_conn_string(
        cls, conn_string: str
    ) -> AsyncIterator[AsyncSqliteSaver]:
        """Create a new AsyncSqliteSaver instance from a connection string.

        Args:
            conn_string (str): The async SQLite connection string. It can have the 'sqlite+aiosqlite:///' prefix.

        Yields
        ------
            AsyncSqliteSaverWithPrefix: A new connected AsyncSqliteSaverWithPrefix instance.
        """
        conn_string = conn_string.split("///")[-1]
        async with super().from_conn_string(conn_string) as memory:
            yield AsyncSqliteSaverWithPrefix(memory.conn)


class AsyncPostgresSaverWithPrefix(AsyncPostgresSaver):
    """Wrapper around the AsyncSqliteSaver that accepts a connection string with prefix."""

    @classmethod
    @asynccontextmanager
    async def from_conn_string(
        cls,
        conn_string: str,
        *,
        pipeline: bool = False,
        serde: SerializerProtocol | None = None,
    ) -> AsyncIterator[AsyncPostgresSaver]:
        """Create a new AsyncPostgresSaver instance from a connection string.

        Args:
            conn_string (str): The async Postgres connection string. It can have the 'postgresql+asyncpg://' prefix.

        Yields
        ------
            AsyncPostgresSaverWithPrefix: A new connected AsyncPostgresSaverWithPrefix instance.
        """
        prefix, body = conn_string.split("://", maxsplit=1)
        currated_prefix = prefix.split("+", maxsplit=1)[0]  # Still works if + not there
        conn_string = currated_prefix + "://" + body
        async with super().from_conn_string(
            conn_string, pipeline=pipeline, serde=serde
        ) as memory:
            yield AsyncPostgresSaverWithPrefix(memory.conn, memory.pipe, memory.serde)



"""Get Morpho tool."""

import logging
from typing import Any, Optional, Type

from langchain_core.tools import ToolException
from pydantic import BaseModel, Field

from neuroagent.cell_types import get_celltypes_descendants
from neuroagent.tools.base_tool import BaseToolOutput, BasicTool
from neuroagent.utils import get_descendants_id

logger = logging.getLogger(__name__)


class InputGetMorpho(BaseModel):
    """Inputs of the knowledge graph API."""

    brain_region_id: str = Field(
        description="ID of the brain region of interest. To get this ID, please use the `resolve-entities-tool` first."
    )
    mtype_id: Optional[str] = Field(
        default=None,
        description="ID of the M-type of interest. To get this ID, please use the `resolve-entities-tool` first.",
    )


class KnowledgeGraphOutput(BaseToolOutput):
    """Output schema for the knowledge graph API."""

    morphology_id: str
    morphology_name: str | None
    morphology_description: str | None
    mtype: str | None

    brain_region_id: str
    brain_region_label: str | None

    subject_species_label: str | None
    subject_age: str | None


class GetMorphoTool(BasicTool):
    """Class defining the Get Morpho logic."""

    name: str = "get-morpho-tool"
    description: str = """Searches a neuroscience based knowledge graph to retrieve neuron morphology names, IDs and descriptions.
    Requires a 'brain_region_id' which is the ID of the brain region of interest as registered in the knowledge graph.
    Optionally accepts an mtype_id.
    The output is a list of morphologies, containing:
    - The brain region ID.
    - The brain region name.
    - The subject species name.
    - The subject age.
    - The morphology ID.
    - The morphology name.
    - the morphology description.
    The morphology ID is in the form of an HTTP(S) link such as 'https://bbp.epfl.ch/neurosciencegraph/data/neuronmorphologies...'."""
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputGetMorpho

    def _run(self) -> None:
        pass

    async def _arun(
        self, brain_region_id: str, mtype_id: str | None = None
    ) -> list[KnowledgeGraphOutput]:
        """From a brain region ID, extract morphologies.

        Parameters
        ----------
        brain_region_id
            ID of the brain region of interest (of the form http://api.brain-map.org/api/v2/data/Structure/...)
        mtype_id
            ID of the mtype of the morphology

        Returns
        -------
            list of KnowledgeGraphOutput to describe the morphology and its metadata, or an error dict.
        """
        logger.info(
            f"Entering Get Morpho tool. Inputs: {brain_region_id=}, {mtype_id=}"
        )
        try:
            # From the brain region ID, get the descendants.
            hierarchy_ids = get_descendants_id(
                brain_region_id, json_path=self.metadata["brainregion_path"]
            )
            logger.info(
                f"Found {len(list(hierarchy_ids))} children of the brain ontology."
            )

            # Create the ES query to query the KG.
            mtype_ids = (
                get_celltypes_descendants(mtype_id, self.metadata["celltypes_path"])
                if mtype_id
                else None
            )
            entire_query = self.create_query(
                brain_regions_ids=hierarchy_ids, mtype_ids=mtype_ids
            )

            # Send the query to get morphologies.
            response = await self.metadata["httpx_client"].post(
                url=self.metadata["url"],
                headers={"Authorization": f"Bearer {self.metadata['token']}"},
                json=entire_query,
            )

            # Process the output and return.
            return self._process_output(response.json())

        except Exception as e:
            raise ToolException(str(e), self.name)

    def create_query(
        self, brain_regions_ids: set[str], mtype_ids: set[str] | None = None
    ) -> dict[str, Any]:
        """Create ES query out of the BR and mtype IDs.

        Parameters
        ----------
        brain_regions_ids
            IDs of the brain region of interest (of the form http://api.brain-map.org/api/v2/data/Structure/...)
        mtype_ids
            IDs the the mtype of the morphology

        Returns
        -------
            dict containing the elasticsearch query to send to the KG.
        """
        # At least one of the children brain region should match.
        conditions = [
            {
                "bool": {
                    "should": [
                        {"term": {"brainRegion.@id.keyword": hierarchy_id}}
                        for hierarchy_id in brain_regions_ids
                    ]
                }
            }
        ]

        if mtype_ids:
            # The correct mtype should match. For now
            # It is a one term should condition, but eventually
            # we will resolve the subclasses of the mtypes.
            # They will all be appended here.
            conditions.append(
                {
                    "bool": {
                        "should": [
                            {"term": {"mType.@id.keyword": mtype_id}}
                            for mtype_id in mtype_ids
                        ]
                    }
                }
            )

        # Assemble the query to return morphologies.
        entire_query = {
            "size": self.metadata["search_size"],
            "track_total_hits": True,
            "query": {
                "bool": {
                    "must": [
                        *conditions,
                        {
                            "term": {
                                "@type.keyword": "https://neuroshapes.org/ReconstructedNeuronMorphology"
                            }
                        },
                        {"term": {"deprecated": False}},
                        {"term": {"curated": True}},
                    ]
                }
            },
        }
        return entire_query

    @staticmethod
    def _process_output(output: Any) -> list[KnowledgeGraphOutput]:
        """Process output to fit the KnowledgeGraphOutput pydantic class defined above.

        Parameters
        ----------
        output
            Raw output of the _arun method, which comes from the KG

        Returns
        -------
            list of KGMorphoFeatureOutput to describe the morphology and its metadata.
        """
        formatted_output = [
            KnowledgeGraphOutput(
                morphology_id=res["_source"]["@id"],
                morphology_name=res["_source"].get("name"),
                morphology_description=res["_source"].get("description"),
                mtype=(
                    res["_source"]["mType"].get("label")
                    if "mType" in res["_source"]
                    else None
                ),
                brain_region_id=res["_source"]["brainRegion"]["@id"],
                brain_region_label=res["_source"]["brainRegion"].get("label"),
                subject_species_label=(
                    res["_source"]["subjectSpecies"].get("label")
                    if "subjectSpecies" in res["_source"]
                    else None
                ),
                subject_age=(
                    res["_source"]["subjectAge"].get("label")
                    if "subjectAge" in res["_source"]
                    else None
                ),
            )
            for res in output["hits"]["hits"]
        ]
        return formatted_output



"""Tool to resolve the brain region from natural english to a KG ID."""

import logging
from typing import Any, Type

from langchain_core.tools import ToolException
from pydantic import BaseModel, Field

from neuroagent.resolving import resolve_query
from neuroagent.tools.base_tool import (
    ETYPE_IDS,
    BaseToolOutput,
    BasicTool,
    EtypesLiteral,
)

logger = logging.getLogger(__name__)


class InputResolveBR(BaseModel):
    """Inputs of the Resolve Brain Region tool.."""

    brain_region: str = Field(
        description="Brain region of interest specified by the user in natural english."
    )
    mtype: str | None = Field(
        default=None,
        description="M-type of interest specified by the user in natural english.",
    )
    etype: EtypesLiteral | None = Field(
        default=None,
        description=(
            "E-type of interest specified by the user in natural english. Possible values:"
            f" {', '.join(list(ETYPE_IDS.keys()))}. The first letter meaning c: continuous,"
            "b: bursting or d: delayed, The other letters in capital meaning AC: accomodating,"
            "NAC: non-accomodating, AD: adapting, NAD: non-adapting, STUT: stuttering,"
            "IR: irregular spiking. Optional suffixes in lowercase can exist:"
            "pyr: pyramidal, int: interneuron, _ltb: low threshold bursting,"
            "_noscltb: non-oscillatory low-threshold bursting. Examples: "
            "cADpyr: continuous adapting pyramidal. dAD_ltb: delayed adapting low-threshold bursting"
        ),
    )


class BRResolveOutput(BaseToolOutput):
    """Output schema for the Brain region resolver."""

    brain_region_name: str
    brain_region_id: str


class MTypeResolveOutput(BaseToolOutput):
    """Output schema for the Mtype resolver."""

    mtype_name: str
    mtype_id: str


class EtypeResolveOutput(BaseToolOutput):
    """Output schema for the Mtype resolver."""

    etype_name: str
    etype_id: str


class ResolveEntitiesTool(BasicTool):
    """Class defining the Brain Region Resolving logic."""

    name: str = "resolve-entities-tool"
    description: str = """From a brain region name written in natural english, search a knowledge graph to retrieve its corresponding ID.
    Optionaly resolve the mtype name from natural english to its corresponding ID too.
    You MUST use this tool when a brain region is specified in natural english because in that case the output of this tool is essential to other tools.
    returns a dictionary containing the brain region name, id and optionaly the mtype name and id.
    Brain region related outputs are stored in the class `BRResolveOutput` while the mtype related outputs are stored in the class `MTypeResolveOutput`."""
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputResolveBR

    def _run(self) -> None:
        """Not implemented yet."""
        pass

    async def _arun(
        self, brain_region: str, mtype: str | None = None, etype: str | None = None
    ) -> list[BRResolveOutput | MTypeResolveOutput | EtypeResolveOutput]:
        """Given a brain region in natural language, resolve its ID.

        Parameters
        ----------
        brain_region
            Name of the brain region to resolve (in english)
        mtype
            Name of the mtype to resolve (in english)

        Returns
        -------
            Mapping from BR/mtype name to ID.
        """
        logger.info(
            f"Entering Brain Region resolver tool. Inputs: {brain_region=}, {mtype=}, {etype=}"
        )
        try:
            # Prepare the output list.
            output: list[BRResolveOutput | MTypeResolveOutput | EtypeResolveOutput] = []

            # First resolve the brain regions.
            brain_regions = await resolve_query(
                sparql_view_url=self.metadata["kg_sparql_url"],
                token=self.metadata["token"],
                query=brain_region,
                resource_type="nsg:BrainRegion",
                search_size=10,
                httpx_client=self.metadata["httpx_client"],
                es_view_url=self.metadata["kg_class_view_url"],
            )
            # Extend the resolved BRs.
            output.extend(
                [
                    BRResolveOutput(
                        brain_region_name=br["label"], brain_region_id=br["id"]
                    )
                    for br in brain_regions
                ]
            )

            # Optionally resolve the mtypes.
            if mtype is not None:
                mtypes = await resolve_query(
                    sparql_view_url=self.metadata["kg_sparql_url"],
                    token=self.metadata["token"],
                    query=mtype,
                    resource_type="bmo:BrainCellType",
                    search_size=10,
                    httpx_client=self.metadata["httpx_client"],
                    es_view_url=self.metadata["kg_class_view_url"],
                )
                # Extend the resolved mtypes.
                output.extend(
                    [
                        MTypeResolveOutput(
                            mtype_name=mtype["label"], mtype_id=mtype["id"]
                        )
                        for mtype in mtypes
                    ]
                )

            # Optionally resolve the etype
            if etype is not None:
                output.append(
                    EtypeResolveOutput(etype_name=etype, etype_id=ETYPE_IDS[etype])
                )

            return output
        except Exception as e:
            raise ToolException(str(e), self.name)



"""Tools folder."""

from neuroagent.tools.bluenaas_tool import BlueNaaSTool
from neuroagent.tools.electrophys_tool import ElectrophysFeatureTool, FeaturesOutput
from neuroagent.tools.get_me_model_tool import GetMEModelTool
from neuroagent.tools.get_morpho_tool import GetMorphoTool, KnowledgeGraphOutput
from neuroagent.tools.kg_morpho_features_tool import (
    KGMorphoFeatureOutput,
    KGMorphoFeatureTool,
)
from neuroagent.tools.literature_search_tool import (
    LiteratureSearchTool,
    ParagraphMetadata,
)
from neuroagent.tools.morphology_features_tool import (
    MorphologyFeatureOutput,
    MorphologyFeatureTool,
)
from neuroagent.tools.resolve_entities_tool import (
    BRResolveOutput,
    ResolveEntitiesTool,
)
from neuroagent.tools.traces_tool import GetTracesTool, TracesOutput

__all__ = [
    "BlueNaaSTool",
    "BRResolveOutput",
    "ElectrophysFeatureTool",
    "FeaturesOutput",
    "GetMorphoTool",
    "GetTracesTool",
    "KGMorphoFeatureOutput",
    "KGMorphoFeatureTool",
    "KnowledgeGraphOutput",
    "LiteratureSearchTool",
    "MorphologyFeatureOutput",
    "MorphologyFeatureTool",
    "ParagraphMetadata",
    "ResolveEntitiesTool",
    "TracesOutput",
    "GetMEModelTool",
]



"""Literature Search tool."""

import logging
from typing import Any, Type

from langchain_core.tools import ToolException
from pydantic import BaseModel, Field

from neuroagent.tools.base_tool import BaseToolOutput, BasicTool

logger = logging.getLogger(__name__)


class InputLiteratureSearch(BaseModel):
    """Inputs of the literature search API."""

    query: str = Field(
        description=(
            "Query to match against the text of paragraphs coming from scientific"
            " articles. The matching is done using the bm25 algorithm, so the query"
            " should be based on keywords to ensure maximal efficiency."
        )
    )


class ParagraphMetadata(BaseToolOutput, extra="ignore"):
    """Metadata for an article."""

    article_title: str
    article_authors: list[str]
    paragraph: str
    section: str | None = None
    article_doi: str | None = None
    journal_issn: str | None = None


class LiteratureSearchTool(BasicTool):
    """Class defining the Literature Search logic."""

    name: str = "literature-search-tool"
    description: str = """Searches the scientific literature. The tool should be used to gather general scientific knowledge. It is best suited for questions about neuroscience and medicine that are not about morphologies.
    It returns a list of paragraphs fron scientific papers that match the query (in the sense of the bm25 algorithm), alongside with the metadata of the articles they were extracted from, such as:
    - title
    - authors
    - paragraph_text
    - section
    - article_doi
    - journal_issn"""
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputLiteratureSearch

    def _run(self, query: str) -> list[ParagraphMetadata]:
        """Search the scientific literature and returns citations.

        Parameters
        ----------
        query
            Query to send to the literature search backend

        Returns
        -------
            List of paragraphs and their metadata
        """
        # Prepare the request's body
        req_body = {
            "query": query,
            "retriever_k": self.metadata["retriever_k"],
            "use_reranker": self.metadata["use_reranker"],
            "reranker_k": self.metadata["reranker_k"],
        }

        # Send the request
        return self._process_output(
            self.metadata["httpx_client"]
            .get(
                self.metadata["url"],
                headers={"Authorization": f"Bearer {self.metadata['token']}"},
                json=req_body,
                timeout=None,
            )
            .json()
        )

    async def _arun(self, query: str) -> list[ParagraphMetadata] | str:
        """Async search the scientific literature and returns citations.

        Parameters
        ----------
        query
            Query to send to the literature search backend

        Returns
        -------
            List of paragraphs and their metadata
        """
        try:
            logger.info(f"Entering literature search tool. Inputs: {query=}")

            # Prepare the request's body
            req_body = {
                "query": query,
                "retriever_k": self.metadata["retriever_k"],
                "use_reranker": self.metadata["use_reranker"],
                "reranker_k": self.metadata["reranker_k"],
            }

            # Send the request
            response = await self.metadata["httpx_client"].get(
                self.metadata["url"],
                headers={"Authorization": f"Bearer {self.metadata['token']}"},
                params=req_body,
                timeout=None,
            )

            return self._process_output(response.json())
        except Exception as e:
            raise ToolException(str(e), self.name)

    @staticmethod
    def _process_output(output: list[dict[str, Any]]) -> list[ParagraphMetadata]:
        """Process output."""
        paragraphs_metadata = [
            ParagraphMetadata(
                article_title=paragraph["article_title"],
                article_authors=paragraph["article_authors"],
                paragraph=paragraph["paragraph"],
                section=paragraph["section"],
                article_doi=paragraph["article_doi"],
                journal_issn=paragraph["journal_issn"],
            )
            for paragraph in output
        ]
        return paragraphs_metadata



"""Cell types metadata."""

import json
import logging
from collections import defaultdict
from pathlib import Path
from typing import Any

logger = logging.getLogger(__file__)


class CellTypesMeta:
    """Class holding the hierarchical cell types metadata.

    Typically, such information would be parsed from a `celltypes.json`
    file.
    """

    def __init__(self) -> None:
        self.name_: dict[str, str] = {}
        self.descendants_ids: dict[str, set[str]] = {}

    def descendants(self, ids: str | set[str]) -> set[str]:
        """Find all descendants of given cell type.

        The result is inclusive, i.e. the input region IDs will be
        included in the result.

        Parameters
        ----------
        ids : set or iterable of set
            A region ID or a collection of region IDs to collect
            descendants for.

        Returns
        -------
        set
            All descendant region IDs of the given regions, including the input cell type themselves.
        """
        if isinstance(ids, str):
            unique_ids = {ids}
        else:
            unique_ids = set(ids)

        descendants = unique_ids.copy()
        for id_ in unique_ids:
            try:
                descendants.update(self.descendants_ids[id_])
            except KeyError:
                logger.info(f"{id_} does not have any child in the hierarchy.")
        return descendants

    def save_config(self, json_file_path: str | Path) -> None:
        """Save the actual configuration in a json file.

        Parameters
        ----------
        json_file_path
            Path where to save the json file
        """
        descendants = {}
        for k, v in self.descendants_ids.items():
            descendants[k] = list(v)

        to_save = {
            "names": self.name_,
            "descendants_ids": descendants,
        }
        with open(json_file_path, "w") as fs:
            fs.write(json.dumps(to_save))

    @classmethod
    def load_config(cls, json_file_path: str | Path) -> "CellTypesMeta":
        """Load a configuration in a json file and return a 'CellTypesMeta' instance.

        Parameters
        ----------
        json_file_path
            Path to the json file containing the brain region hierarchy

        Returns
        -------
            RegionMeta class with pre-loaded hierarchy
        """
        with open(json_file_path, "r") as fs:
            to_load = json.load(fs)

        descendants_ids = {}
        for k, v in to_load["descendants_ids"].items():
            descendants_ids[k] = set(v)

        self = cls()

        self.name_ = to_load["names"]
        self.descendants_ids = descendants_ids
        return self

    @classmethod
    def from_dict(cls, hierarchy: dict[str, Any]) -> "CellTypesMeta":
        """Load the structure graph from a dict and create a Class instance.

        Parameters
        ----------
        hierarchy : dict[str, Any]
            Hierarchy in dictionary format.

        Returns
        -------
        RegionMeta
            The initialized instance of this class.
        """
        names = {}
        initial_json: dict[str, set[str]] = defaultdict(set)
        for i in range(len(hierarchy["defines"])):
            cell_type = hierarchy["defines"][i]
            names[cell_type["@id"]] = (
                cell_type["label"] if "label" in cell_type else None
            )
            if "subClassOf" not in cell_type.keys():
                initial_json[cell_type["@id"]] = set()
                continue
            parents = cell_type["subClassOf"]
            for parent in parents:
                initial_json[parent].add(hierarchy["defines"][i]["@id"])

        current_json = initial_json.copy()

        for i in range(10):  # maximum number of attempts
            new_json = {}
            for k, v in current_json.items():
                new_set = v.copy()
                for child in v:
                    if child in current_json.keys():
                        new_set.update(current_json[child])
                new_json[k] = new_set

            if new_json == current_json:
                break

            if i == 9:
                raise ValueError("Did not manage to create a CellTypesMeta object.")

            current_json = new_json.copy()

        self = cls()

        self.name_ = names
        self.descendants_ids = new_json

        return self

    @classmethod
    def from_json(cls, json_path: Path | str) -> "CellTypesMeta":
        """Load the structure graph from a JSON file and create a Class instance.

        Parameters
        ----------
        json_path : str or pathlib.Path

        Returns
        -------
        RegionMeta
            The initialized instance of this class.
        """
        with open(json_path) as fh:
            hierarchy = json.load(fh)

        return cls.from_dict(hierarchy)


def get_celltypes_descendants(cell_type_id: str, json_path: str | Path) -> set[str]:
    """Get all descendant of a brain region id.

    Parameters
    ----------
    cell_type_id
        Cell type ID for which to find the descendants list.
    json_path
        Path to the json file containing the Cell Types hierarchy.

    Returns
    -------
        Set of descendants of a cell type
    """
    try:
        region_meta = CellTypesMeta.load_config(json_path)
        hierarchy_ids = region_meta.descendants(cell_type_id)
    except IOError:
        logger.warning(f"The file {json_path} doesn't exist.")
        hierarchy_ids = {cell_type_id}

    return hierarchy_ids



"""Simple agent."""

import logging
from contextlib import AsyncExitStack
from typing import Any, AsyncIterator

from langchain_core.messages import AIMessage, HumanMessage
from langgraph.checkpoint.base import BaseCheckpointSaver
from langgraph.prebuilt import create_react_agent
from pydantic import model_validator

from neuroagent.agents import AgentOutput, AgentStep, BaseAgent

logger = logging.getLogger(__name__)


class SimpleChatAgent(BaseAgent):
    """Simple Agent class."""

    memory: BaseCheckpointSaver[Any]

    @model_validator(mode="before")
    @classmethod
    def create_agent(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Instantiate the clients upon class creation."""
        data["agent"] = create_react_agent(
            model=data["llm"],
            tools=data["tools"],
            checkpointer=data["memory"],
            state_modifier="""You are a helpful assistant helping scientists with neuro-scientific questions.
                You must always specify in your answers from which brain regions the information is extracted.
                Do no blindly repeat the brain region requested by the user, use the output of the tools instead.""",
        )
        return data

    def run(self, session_id: str, query: str) -> Any:
        """Run the agent against a query."""
        pass

    async def arun(self, thread_id: str, query: str) -> Any:
        """Run the agent against a query."""
        config = {"configurable": {"thread_id": thread_id}}
        input_message = HumanMessage(content=query)
        result = await self.agent.ainvoke({"messages": [input_message]}, config=config)
        return self._process_output(result)

    async def astream(
        self, thread_id: str, query: str, connection_string: str | None = None
    ) -> AsyncIterator[str]:
        """Run the agent against a query in streaming way.

        Parameters
        ----------
        thread_id
            ID of the thread of the chat.
        query
            Query of the user.
        connection_string
            connection string for the checkpoint database.

        Yields
        ------
            Iterator streaming the processed output of the LLM
        """
        async with (
            self.agent.checkpointer.__class__.from_conn_string(connection_string)
            if connection_string
            else AsyncExitStack() as memory
        ):
            if isinstance(memory, BaseCheckpointSaver):
                self.agent.checkpointer = memory
            config = {"configurable": {"thread_id": thread_id}}
            streamed_response = self.agent.astream_events(
                {"messages": query}, version="v2", config=config
            )
            is_streaming = False
            async for event in streamed_response:
                kind = event["event"]
                # check for the model stream.
                if kind == "on_chat_model_stream":
                    # check if we are calling the tools.
                    data_chunk = event["data"]["chunk"]
                    if "tool_calls" in data_chunk.additional_kwargs:
                        tool = data_chunk.additional_kwargs["tool_calls"]
                        if tool[0]["function"]["name"]:
                            yield (
                                f'\nCalling tool : {tool[0]["function"]["name"]} with'
                                " arguments : "
                            )
                        if tool[0]["function"]["arguments"]:
                            yield tool[0]["function"]["arguments"]

                    content = data_chunk.content
                    if content:
                        if not is_streaming:
                            yield "\n<begin_llm_response>\n"
                        is_streaming = True
                        yield content
            yield "\n"

    @staticmethod
    def _process_output(output: Any) -> AgentOutput:
        """Format the output.

        Parameters
        ----------
        output
            Raw output of the LLM

        Returns
        -------
            Unified output across different agent type.
        """
        # Gather tool name and arguments together
        agent_steps = [
            AgentStep(
                tool_name=tool_call["name"],
                arguments=tool_call["args"],
            )
            for step in output["messages"]
            if isinstance(step, AIMessage) and step.additional_kwargs
            for tool_call in step.tool_calls
        ]
        return AgentOutput(response=output["messages"][-1].content, steps=agent_steps)



"""Test cell types meta functions."""

from pathlib import Path

import pytest

from neuroagent.cell_types import CellTypesMeta, get_celltypes_descendants

CELL_TYPES_FILE = Path(__file__).parent / "data" / "kg_cell_types_hierarchy_test.json"


@pytest.mark.parametrize(
    "cell_type_id,expected_descendants",
    [
        (
            "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType",
            {
                "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/L23_PTPC",
                "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/GCL_GC",
                "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType",
            },
        ),
        (
            "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
            {
                "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
            },
        ),
        (
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
            {
                "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
            },
        ),
    ],
)
def test_get_celltypes_descendants(cell_type_id, expected_descendants, tmp_path):
    cell_types_meta = CellTypesMeta.from_json(CELL_TYPES_FILE)
    save_file = tmp_path / "tmp_config_cell_types_meta.json"
    cell_types_meta.save_config(save_file)

    descendants = get_celltypes_descendants(cell_type_id, json_path=save_file)
    assert expected_descendants == descendants


class TestCellTypesMeta:
    def test_from_json(self):
        ct_meta = CellTypesMeta.from_json(CELL_TYPES_FILE)
        assert isinstance(ct_meta.name_, dict)
        assert isinstance(ct_meta.descendants_ids, dict)

        expected_names = {
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint": "cACint",
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/GCL_GC": "GCL_GC",
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/L23_PTPC": (
                "L23_PTPC"
            ),
        }

        assert ct_meta.name_ == expected_names
        assert ct_meta.descendants_ids[
            "https://bbp.epfl.ch/ontologies/core/mtypes/HippocampusMType"
        ] == {"http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/GCL_GC"}
        assert ct_meta.descendants_ids[
            "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType"
        ] == {
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/L23_PTPC",
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
            "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/GCL_GC",
        }

    @pytest.mark.parametrize(
        "cell_type_id,expected_descendants",
        [
            (
                "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType",
                {
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/L23_PTPC",
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/GCL_GC",
                    "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType",
                },
            ),
            (
                "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                {
                    "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                },
            ),
            (
                [
                    "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType",
                    "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                ],
                {
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/L23_PTPC",
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/mtypes/GCL_GC",
                    "https://bbp.epfl.ch/ontologies/core/bmo/BrainCellType",
                    "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                },
            ),
            (
                "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                {
                    "https://bbp.epfl.ch/ontologies/core/bmo/NeuronElectricalType",
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                },
            ),
            (
                "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                {
                    "http://bbp.epfl.ch/neurosciencegraph/ontologies/etypes/cACint",
                },
            ),
        ],
    )
    def test_descendants(self, cell_type_id, expected_descendants):
        ct_meta = CellTypesMeta.from_json(CELL_TYPES_FILE)
        assert ct_meta.descendants(cell_type_id) == expected_descendants

    def test_load_and_save_config(self, tmp_path):
        ct_meta = CellTypesMeta.from_json(CELL_TYPES_FILE)
        file_path = tmp_path / "ct_meta_tmp.json"
        ct_meta.save_config(file_path)
        ct_meta2 = CellTypesMeta.load_config(file_path)
        assert ct_meta.name_ == ct_meta2.name_
        assert ct_meta.descendants_ids == ct_meta2.descendants_ids



"""Electrophys tool."""

import logging
import tempfile
from statistics import mean
from typing import Any, Literal, Optional, Type

from bluepyefe.extract import extract_efeatures
from efel.units import get_unit
from langchain_core.tools import ToolException
from pydantic import BaseModel, Field

from neuroagent.tools.base_tool import BaseToolOutput, BasicTool
from neuroagent.utils import get_kg_data

logger = logging.getLogger(__name__)


POSSIBLE_PROTOCOLS = {
    "idrest": ["idrest"],
    "idthresh": ["idthres", "idthresh"],
    "iv": ["iv"],
    "apwaveform": ["apwaveform"],
    "spontaneous": ["spontaneous"],
    "step": ["step"],
    "spontaps": ["spontaps"],
    "firepattern": ["firepattern"],
    "sponnohold30": ["sponnohold30", "spontnohold30"],
    "sponthold30": ["sponhold30", "sponthold30"],
    "starthold": ["starthold"],
    "startnohold": ["startnohold"],
    "delta": ["delta"],
    "sahp": ["sahp"],
    "idhyperpol": ["idhyeperpol"],
    "irdepol": ["irdepol"],
    "irhyperpol": ["irhyperpol"],
    "iddepol": ["iddepol"],
    "ramp": ["ramp"],
    "apthresh": ["apthresh", "ap_thresh"],
    "hyperdepol": ["hyperdepol"],
    "negcheops": ["negcheops"],
    "poscheops": ["poscheops"],
    "spikerec": ["spikerec"],
    "sinespec": ["sinespec"],
}


STIMULI_TYPES = list[
    Literal[
        "spontaneous",
        "idrest",
        "idthres",
        "apwaveform",
        "iv",
        "step",
        "spontaps",
        "firepattern",
        "sponnohold30",
        "sponhold30",
        "starthold",
        "startnohold",
        "delta",
        "sahp",
        "idhyperpol",
        "irdepol",
        "irhyperpol",
        "iddepol",
        "ramp",
        "ap_thresh",
        "hyperdepol",
        "negcheops",
        "poscheops",
        "spikerec",
        "sinespec",
    ]
]

CALCULATED_FEATURES = list[
    Literal[
        "spike_count",
        "time_to_first_spike",
        "time_to_last_spike",
        "inv_time_to_first_spike",
        "doublet_ISI",
        "inv_first_ISI",
        "ISI_log_slope",
        "ISI_CV",
        "irregularity_index",
        "adaptation_index",
        "mean_frequency",
        "strict_burst_number",
        "strict_burst_mean_freq",
        "spikes_per_burst",
        "AP_height",
        "AP_amplitude",
        "AP1_amp",
        "APlast_amp",
        "AP_duration_half_width",
        "AHP_depth",
        "AHP_time_from_peak",
        "AP_peak_upstroke",
        "AP_peak_downstroke",
        "voltage_base",
        "voltage_after_stim",
        "ohmic_input_resistance_vb_ssse",
        "steady_state_voltage_stimend",
        "sag_amplitude",
        "decay_time_constant_after_stim",
        "depol_block_bool",
    ]
]


class AmplitudeInput(BaseModel):
    """Amplitude class."""

    min_value: float
    max_value: float


class InputElectrophys(BaseModel):
    """Inputs of the NeuroM API."""

    trace_id: str = Field(
        description=(
            "ID of the trace of interest. The trace ID is in the form of an HTTP(S)"
            " link such as 'https://bbp.epfl.ch/neurosciencegraph/data/traces...'."
        )
    )
    stimuli_types: Optional[STIMULI_TYPES] = Field(
        description=(
            "Type of stimuli requested by the user. Should be one of 'spontaneous',"
            " 'idrest', 'idthres', 'apwaveform', 'iv', 'step', 'spontaps',"
            " 'firepattern', 'sponnohold30','sponhold30', 'starthold', 'startnohold',"
            " 'delta', 'sahp', 'idhyperpol', 'irdepol', 'irhyperpol','iddepol', 'ramp',"
            " 'ap_thresh', 'hyperdepol', 'negcheops', 'poscheops',"
            " 'spikerec', 'sinespec'."
        )
    )
    calculated_feature: Optional[CALCULATED_FEATURES] = Field(
        description=(
            "Feature requested by the user. Should be one of 'spike_count',"
            "'time_to_first_spike', 'time_to_last_spike',"
            "'inv_time_to_first_spike', 'doublet_ISI', 'inv_first_ISI',"
            "'ISI_log_slope', 'ISI_CV', 'irregularity_index', 'adaptation_index',"
            "'mean_frequency', 'strict_burst_number', 'strict_burst_mean_freq',"
            "'spikes_per_burst', 'AP_height', 'AP_amplitude', 'AP1_amp', 'APlast_amp',"
            "'AP_duration_half_width', 'AHP_depth', 'AHP_time_from_peak',"
            "'AP_peak_upstroke', 'AP_peak_downstroke', 'voltage_base',"
            "'voltage_after_stim', 'ohmic_input_resistance_vb_ssse',"
            "'steady_state_voltage_stimend', 'sag_amplitude',"
            "'decay_time_constant_after_stim', 'depol_block_bool'"
        )
    )
    amplitude: Optional[AmplitudeInput] = Field(
        description=(
            "Amplitude of the protocol (should be specified in nA)."
            "Can be a range of amplitudes with min and max values"
            "Can be None (if the user does not specify it)"
            " and all the amplitudes are going to be taken into account."
        ),
    )


class FeaturesOutput(BaseToolOutput):
    """Output schema for the neurom tool."""

    brain_region: str
    feature_dict: dict[str, Any]


class ElectrophysFeatureTool(BasicTool):
    """Class defining the Electrophys Featyres Tool."""

    name: str = "electrophys-features-tool"
    description: str = """Given a trace ID, extract features from the trace for certain stimuli types and certain amplitudes.
    You can optionally specify which feature to calculate, for which stimuli types and for which amplitudes:
    - The calculated features are a list of features that the user requests to compute.
    - The stimuli types are the types of input stimuli injected in the cell when measuring the response.
    - The amplitude is the total current injected in the cell when measuring the response.
    Specify those ONLY if the user specified them. Otherwise leave them as None.
    """
    metadata: dict[str, Any]
    args_schema: Type[BaseModel] = InputElectrophys

    def _run(self) -> None:  # type: ignore
        """Not implemented yet."""
        pass

    async def _arun(
        self,
        trace_id: str,
        calculated_feature: CALCULATED_FEATURES | None = None,
        stimuli_types: STIMULI_TYPES | None = None,
        amplitude: AmplitudeInput | None = None,
    ) -> FeaturesOutput | dict[str, str]:
        """Give features about trace.

        Parameters
        ----------
        trace
            ID of the trace of interest (of the form https://bbp.epfl.ch/neurosciencegraph/data/traces...)
        calculated_features
            List of features one wants to compute
        stimuli_types
            List of stimuli types that should be taken into account when computing the features
        amplitude
            Amplitude range of the input stimulus when measuring the cell's response

        Returns
        -------
            Dict of feature: value
        """
        logger.info(
            f"Entering electrophys tool. Inputs: {trace_id=}, {calculated_feature=},"
            f" {amplitude=}, {stimuli_types=}"
        )
        try:
            # Deal with cases where user did not specify stimulus type or/and feature
            if not stimuli_types:
                # Default to IDRest if protocol not specified
                logger.warning("No stimulus type specified. Defaulting to IDRest.")
                stimuli_types = ["idrest"]
            if not calculated_feature:
                # Compute ALL of the available features if not specified
                logger.warning("No feature specified. Defaulting to everything.")
                calculated_feature = list(CALCULATED_FEATURES.__args__[0].__args__)  # type: ignore

            # Download the .nwb file associated to the trace from the KG
            trace_content, metadata = await get_kg_data(
                object_id=trace_id,
                httpx_client=self.metadata["httpx_client"],
                url=self.metadata["url"],
                token=self.metadata["token"],
                preferred_format="nwb",
            )

            # Turn amplitude requirement of user into a bluepyefe compatible representation
            if isinstance(amplitude, AmplitudeInput):
                # If the user specified amplitude/a range of amplitudes,
                # the target amplitude is centered on the range and the
                # tolerance is set as half the range
                desired_amplitude = mean([amplitude.min_value, amplitude.max_value])

                # If the range is just one number, use 10% of it as tolerance
                if amplitude.min_value == amplitude.max_value:
                    desired_tolerance = amplitude.max_value * 0.1
                else:
                    desired_tolerance = amplitude.max_value - desired_amplitude
            else:
                # If the amplitudes are not specified, take an arbitrarily high tolerance
                desired_amplitude = 0
                desired_tolerance = 1e12
            logger.info(
                f"target amplitude set to {desired_amplitude} nA. Tolerance is"
                f" {desired_tolerance} nA"
            )

            targets = []
            # Create a target for each stimuli_types and their various spellings and for each feature to compute
            for stim_type in stimuli_types:
                for efeature in calculated_feature:
                    for protocol in POSSIBLE_PROTOCOLS[stim_type]:
                        target = {
                            "efeature": efeature,
                            "protocol": protocol,
                            "amplitude": desired_amplitude,
                            "tolerance": desired_tolerance,
                        }
                        targets.append(target)
            logger.info(f"Generated {len(targets)} targets.")

            # The trace needs to be opened from a file, no way to hack it
            with (
                tempfile.NamedTemporaryFile(suffix=".nwb") as temp_file,
                tempfile.TemporaryDirectory() as temp_dir,
            ):
                temp_file.write(trace_content)

                # LNMC traces need to be adjusted by an output voltage of 14mV due to their experimental protocol
                if metadata.is_lnmc:
                    files_metadata = {
                        "test": {
                            stim_type: [
                                {
                                    "filepath": temp_file.name,
                                    "protocol": protocol,
                                    "ljp": 14,
                                }
                                for protocol in POSSIBLE_PROTOCOLS[stim_type]
                            ]
                            for stim_type in stimuli_types
                        }
                    }
                else:
                    files_metadata = {
                        "test": {
                            stim_type: [
                                {"filepath": temp_file.name, "protocol": protocol}
                                for protocol in POSSIBLE_PROTOCOLS[stim_type]
                            ]
                            for stim_type in stimuli_types
                        }
                    }

                # Extract the requested features for the requested protocols
                efeatures, protocol_definitions, _ = extract_efeatures(
                    output_directory=temp_dir,
                    files_metadata=files_metadata,
                    targets=targets,
                    absolute_amplitude=True,
                )
                output_features = {}

                # Format the extracted features into a readable dict for the model
                for protocol_name in protocol_definitions.keys():
                    efeatures_values = efeatures[protocol_name]
                    protocol_def = protocol_definitions[protocol_name]
                    output_features[protocol_name] = {
                        f"{f['efeature_name']} (avg on n={f['n']} trace(s))": (
                            f"{f['val'][0]} {get_unit(f['efeature_name']) if get_unit(f['efeature_name']) != 'constant' else ''}".strip()
                        )
                        for f in efeatures_values["soma"]
                    }

                    # Add the stimulus current of the protocol to the output
                    output_features[protocol_name]["stimulus_current"] = (
                        f"{protocol_def['step']['amp']} nA"
                    )
            return FeaturesOutput(
                brain_region=metadata.brain_region, feature_dict=output_features
            )
        except Exception as e:
            raise ToolException(str(e), self.name)
